# encoding: utf-8

#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import tensorflow as tf

from skflow.preprocessing import categorical_vocabulary


class CategoricalVocabularyTest(tf.test.TestCase):

    def testIntVocabulary(self):
        vocab = categorical_vocabulary.CategoricalVocabulary()
        self.assertEqual(vocab.get(1), 1)
        self.assertEqual(vocab.get(3), 2)
        self.assertEqual(vocab.get(2), 3)
        self.assertEqual(vocab.get(3), 2)
        # This vocab doesn't handle nan specially.
        self.assertEqual(vocab.get(float('nan')), 4)
        self.assertEqual(len(vocab), 5)

    def testWordVocabulary(self):
        vocab = categorical_vocabulary.CategoricalVocabulary()
        self.assertEqual(vocab.get('a'), 1)
        self.assertEqual(vocab.get('b'), 2)
        self.assertEqual(vocab.get('a'), 1)
        self.assertEqual(vocab.get('b'), 2)

    def testCountsTrim(self):
        vocab = categorical_vocabulary.CategoricalVocabulary()
        vocab.get('a')
        vocab.add('a', 10)
        vocab.get('c')
        vocab.add('c', 5)
        # not in vocab yet, skips.
        vocab.add('b', 5)
        vocab.trim(7)
        vocab.freeze()
        self.assertEqual(vocab.get('b'), 0)
        self.assertEqual(vocab.get('c'), 0)
        self.assertEqual(len(vocab), 2)


if __name__ == "__main__":
    tf.test.main()



#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import random

import numpy as np

from sklearn import datasets
from sklearn.metrics import accuracy_score, mean_squared_error

import tensorflow as tf

import skflow

class MultiOutputTest(tf.test.TestCase):

    def testMultiRegression(self):
        random.seed(42)
        rng = np.random.RandomState(1)
        X = np.sort(200 * rng.rand(100, 1) - 100, axis=0)
        y = np.array([np.pi * np.sin(X).ravel(), np.pi * np.cos(X).ravel()]).T
        regressor = skflow.TensorFlowLinearRegressor(learning_rate=0.01)
        regressor.fit(X, y)
        score = mean_squared_error(regressor.predict(X), y)
        self.assertLess(score, 10, "Failed with score = {0}".format(score))


if __name__ == "__main__":
    tf.test.main()



"""Scikit Flow Estimators."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from skflow.estimators.base import TensorFlowEstimator
from skflow.estimators.linear import TensorFlowLinearClassifier
from skflow.estimators.linear import TensorFlowClassifier
from skflow.estimators.linear import TensorFlowLinearRegressor
from skflow.estimators.linear import TensorFlowRegressor
from skflow.estimators.dnn import TensorFlowDNNClassifier
from skflow.estimators.dnn import TensorFlowDNNRegressor
from skflow.estimators.rnn import TensorFlowRNNClassifier
from skflow.estimators.rnn import TensorFlowRNNRegressor



"""Methods to allow pandas.DataFrame."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

try:
    import pandas as pd
    HAS_PANDAS = True
except ImportError:
    HAS_PANDAS = False

PANDAS_DTYPES = {'int8': 'int', 'int16': 'int', 'int32': 'int', 'int64': 'int',\
'uint8': 'int', 'uint16': 'int', 'uint32': 'int', 'uint64': 'int', 'float16': 'float',\
'float32': 'float', 'float64': 'float', 'bool': 'i'}


def extract_pandas_data(data):
    """Extract data from pandas.DataFrame for predictors"""
    if not isinstance(data, pd.DataFrame):
        return data

    if all(dtype.name in PANDAS_DTYPES for dtype in data.dtypes):
        return data.values.astype('float')
    else:
        raise ValueError('Data types for data must be int, float, or bool.')


def extract_pandas_matrix(data):
    """Extracts numpy matrix from pandas DataFrame."""
    if not isinstance(data, pd.DataFrame):
        return data

    return data.as_matrix()


def extract_pandas_labels(labels):
    """Extract data from pandas.DataFrame for labels"""
    if isinstance(labels, pd.DataFrame): # pandas.Series also belongs to DataFrame
        if len(labels.columns) > 1:
            raise ValueError('Only one column for labels is allowed.')

        if all(dtype.name in PANDAS_DTYPES for dtype in labels.dtypes):
            return labels.values.astype('float')
        else:
            raise ValueError('Data types for labels must be int, float, or bool.')
    else:
        return labels



"""Categorical vocabulary classes to map categories to indexes.

Can be used for categorical variables, sparse variables and words.
"""

#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import collections
import six


class CategoricalVocabulary(object):
    """Categorical variables vocabulary class.

    Accumulates and provides mapping from classes to indexes.
    Can be easily used for words.
    """

    def __init__(self, unknown_token='<UNK>', support_reverse=True):
        self._mapping = {unknown_token: 0}
        self._support_reverse = support_reverse
        if support_reverse:
            self._reverse_mapping = [unknown_token]
        self._freq = collections.defaultdict(int)
        self._freeze = False

    def __len__(self):
        """Returns total count of mappings. Including unknown token."""
        return len(self._mapping)

    def freeze(self, freeze=True):
        """Freezes the vocabulary, after which new words return unknown token id.

        Args:
            freeze: True to freeze, False to unfreeze.
        """
        self._freeze = freeze

    def get(self, category):
        """Returns word's id in the vocabulary.

        If category is new, creates a new id for it.

        Args:
            category: string or integer to lookup in vocabulary.

        Returns:
            interger, id in the vocabulary.
        """
        if category not in self._mapping:
            if self._freeze:
                return 0
            self._mapping[category] = len(self._mapping)
            if self._support_reverse:
                self._reverse_mapping.append(category)
        return self._mapping[category]

    def add(self, category, count=1):
        """Adds count of the category to the frequency table.

        Args:
            category: string or integer, category to add frequency to.
            count: optional integer, how many to add.
        """
        category_id = self.get(category)
        if category_id <= 0:
            return
        self._freq[category] += count

    def trim(self, min_frequency, max_frequency=-1):
        """Trims vocabulary for minimum frequency.

        Args:
            min_frequency: minimum frequency to keep.
            max_frequency: optional, maximum frequency to keep.
                Useful to remove very frequent categories (like stop words).
        """
        for category, count in six.iteritems(self._freq):
            if count <= min_frequency and (max_frequency < 0 or
                                           count >= max_frequency):
                self._mapping.pop(category)

    def reverse(self, class_id):
        """Given class id reverse to original class name.

        Args:
            class_id: Id of the class.

        Returns:
            Class name.

        Raises:
            ValueError if this vocabulary wasn't initalized with
            support_reverse.
        """
        if not self._support_reverse:
            raise ValueError("This vocabulary wasn't initalized with "
                             "support_reverse to support reverse() function.")
        return self._reverse_mapping[class_id]




"""TensorFlow Ops to work with embeddings.

Note: categorical variables are handled via embeddings in many cases.
For example, in case of words.
"""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function, absolute_import

import tensorflow as tf


def embedding_lookup(params, ids, name="embedding_lookup"):
    """Provides a N dimensional version of tf.embedding_lookup.

    Ids are flattened to a 1d tensor before being passed to embedding_lookup
    then, they are unflattend to match the original ids shape plus an extra
    leading dimension of the size of the embeddings.

    Args:
        params: List of tensors of size D0 x D1 x ... x Dn-2 x Dn-1.
        ids: N-dimensional tensor of B0 x B1 x .. x Bn-2 x Bn-1.
             Must contain indexes into params.
        name: Optional name for the op.

    Returns:
        A tensor of size B0 x B1 x .. x Bn-2 x Bn-1 x D1 x ... x Dn-2 x Dn-1
        containing the values from the params tensor(s) for indecies in ids.

    Raises:
        ValueError: if some parameters are invalid.
    """
    with tf.op_scope([params, ids], name, "embedding_lookup"):
        params = tf.convert_to_tensor(params)
        ids = tf.convert_to_tensor(ids)
        shape = tf.shape(ids)
        ids_flat = tf.reshape(ids, tf.reduce_prod(shape, keep_dims=True))
        embeds_flat = tf.nn.embedding_lookup(params, ids_flat, name)
        embed_shape = tf.concat(0, [shape, [-1]])
        embeds = tf.reshape(embeds_flat, embed_shape)
        embeds.set_shape(ids.get_shape().concatenate(params.get_shape()[1:]))
        return embeds


def categorical_variable(tensor_in, n_classes, embedding_size, name):
    """Creates an embedding for categorical variable with given number of
    classes.

    Args:
        tensor_in: Input tensor with class identifier (can be batch or
            N-dimensional).
        n_classes: Number of classes.
        embedding_size: Size of embedding vector to represent each class.
        name: Name of this categorical variable.
    Returns:
        Tensor of input shape, with additional dimension for embedding.

    Example:
        Calling categorical_variable([1, 2], 5, 10, "my_cat"), will return 2 x 10
        tensor, where each row is representation of the class.
    """
    with tf.variable_scope(name):
        embeddings = tf.get_variable(
            name + "_embeddings", [n_classes, embedding_size])
        return embedding_lookup(embeddings, tensor_in)



"""TensorFlow ops for array / tensor manipulation."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function, absolute_import

import tensorflow as tf


def split_squeeze(dim, num_split, tensor_in):
    """Splits input on given dimension and then squeezes that dimension.

    Args:
        dim: Dimension to split and squeeze on.
        num_split: integer, the number of ways to split.
        tensor_in: Input tensor of shape [N1, N2, .. Ndim, .. Nx].

    Returns:
        List of tensors [N1, N2, .. Ndim-1, Ndim+1, .. Nx].
    """
    return [tf.squeeze(t, squeeze_dims=[dim]) for t in tf.split(dim, num_split, tensor_in)]


def expand_concat(dim, inputs):
    """Expands inputs on given dimension and then concatenates them.

    Args:
        dim: Dimension to expand and concatenate on.
        inputs: List of tensors of the same shape [N1, ... Nx].

    Returns:
        A tensor of shape [N1, .. Ndim, ... Nx]
    """
    return tf.concat(dim, [tf.expand_dims(t, dim) for t in inputs])


def one_hot_matrix(tensor_in, num_classes, on_value=1.0, off_value=0.0):
    """Encodes indices from given tensor as one-hot tensor.

    TODO(ilblackdragon): Ideally implementation should be
    part of TensorFlow with Eigen-native operation.

    Args:
        tensor_in: Input tensor of shape [N1, N2].
        num_classes: Number of classes to expand index into.
        on_value: Tensor or float, value to fill-in given index.
        off_value: Tensor or float, value to fill-in everything else.
    Returns:
        Tensor of shape [N1, N2, num_classes] with 1.0 for each id in original
        tensor.
    """
    tensor_in = tf.convert_to_tensor(tensor_in)
    sparse_values = tf.to_int64(tf.reshape(tensor_in, [-1, 1]))
    size = tf.shape(sparse_values)[0]
    dims = tf.shape(tensor_in)
    indices = tf.to_int64(tf.reshape(tf.range(0, size), [-1, 1]))
    indices_values = tf.concat(1, [indices, sparse_values])
    outshape = tf.to_int64(expand_concat(0, [size, num_classes]))
    one_hot_vector = tf.sparse_to_dense(indices_values, outshape, on_value, off_value)
    ret = tf.reshape(one_hot_vector, tf.concat(0, [dims, [num_classes]]))
    ret.set_shape(tensor_in.get_shape().concatenate(num_classes))
    return ret



"""Dropout operations and handling."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function, absolute_import

import tensorflow as tf


# Key to collect dropout probabilities.
DROPOUTS = "dropouts"


def dropout(tensor_in, prob, name=None):
    """Adds dropout node and stores probability tensor into graph collection.

    Args:
        tensor_in: Input tensor.
        prob: Float or Tensor.

    Returns:
        Tensor of the same shape of `tensor_in`.

    Raises:
        ValueError: If `keep_prob` is not in `(0, 1]`.
    """
    with tf.op_scope([tensor_in], name, "dropout") as name:
        if isinstance(prob, float):
            prob = tf.get_variable("prob", [],
                                   initializer=tf.constant_initializer(prob),
                                   trainable=False)
        tf.add_to_collection(DROPOUTS, prob)
        return tf.nn.dropout(tensor_in, prob)




# encoding: utf-8

#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import unicode_literals

import tensorflow as tf

from skflow.preprocessing import text
from skflow.preprocessing import CategoricalVocabulary


class TextTest(tf.test.TestCase):

    def testTokenizer(self):
        words = text.tokenizer(["a b c", "a\nb\nc", "a, b - c",
                                "фыв выф", "你好 怎么样"])
        self.assertEqual(list(words),
                         [["a", "b", "c"],
                          ["a", "b", "c"],
                          ["a", "b", "-", "c"],
                          ["фыв", "выф"],
                          ["你好", "怎么样"]])

    def testByteProcessor(self):
        processor = text.ByteProcessor(max_document_length=8)
        inp = ["abc", "фыва", "фыва", b"abc",
               "12345678901234567890"]
        res = list(processor.fit_transform(inp))
        self.assertAllEqual(res,
                            [[97, 98, 99, 0, 0, 0, 0, 0],
                            [209, 132, 209, 139, 208, 178, 208, 176],
                            [209, 132, 209, 139, 208, 178, 208, 176],
                            [97, 98, 99, 0, 0, 0, 0, 0],
                            [49, 50, 51, 52, 53, 54, 55, 56]])
        res = list(processor.reverse(res))
        self.assertAllEqual(res,
            ["abc", "фыва", "фыва", "abc", "12345678"])

    def testVocabularyProcessor(self):
        vocab_processor = text.VocabularyProcessor(
            max_document_length=4,
            min_frequency=1)
        tokens = vocab_processor.fit_transform(
            ["a b c", "a\nb\nc", "a, b - c"])
        self.assertAllEqual(list(tokens),
                            [[1, 2, 3, 0], [1, 2, 3, 0], [1, 2, 0, 3]])

    def testExistingVocabularyProcessor(self):
        vocab = CategoricalVocabulary()
        vocab.get("A")
        vocab.get("B")
        vocab.freeze()
        vocab_processor = text.VocabularyProcessor(
            max_document_length=4, vocabulary=vocab, tokenizer_fn=list)
        tokens = vocab_processor.fit_transform(["ABC", "CBABAF"])
        self.assertAllEqual(list(tokens), [[1, 2, 0, 0], [0, 2, 1, 2]])


if __name__ == "__main__":
    tf.test.main()



"""TensorFlow ops for Batch Normalization."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function, absolute_import

import tensorflow as tf


def batch_normalize(tensor_in, epsilon=1e-5, convnet=True, decay=0.9,
                    scale_after_normalization=True):
    """Batch Normalization

    Args:
        tensor_in: input Tensor, 4D shape:
                   [batch, in_height, in_width, in_depth].
        epsilon : A float number to avoid being divided by 0.
        decay: decay rate for exponential moving average.
        convnet: Whether this is for convolutional net use. If this is True,
                 moments will sum across axis [0, 1, 2]. Otherwise, only [0].
        scale_after_normalization: Whether to scale after normalization.
    """
    shape = tensor_in.get_shape().as_list()

    with tf.variable_scope("batch_norm"):
        gamma = tf.get_variable("gamma", [shape[-1]],
                                initializer=tf.random_normal_initializer(1., 0.02))
        beta = tf.get_variable("beta", [shape[-1]],
                               initializer=tf.constant_initializer(0.))
        ema = tf.train.ExponentialMovingAverage(decay=decay)
        if convnet:
            assign_mean, assign_var = tf.nn.moments(tensor_in, [0, 1, 2])
        else:
            assign_mean, assign_var = tf.nn.moments(tensor_in, [0])
        ema_assign_op = ema.apply([assign_mean, assign_var])
        ema_mean, ema_var = ema.average(assign_mean), ema.average(assign_var)
        def update_mean_var():
            """Internal function that updates mean and variance during training"""
            with tf.control_dependencies([ema_assign_op]):
                return tf.identity(assign_mean), tf.identity(assign_var)
        is_training = tf.squeeze(tf.get_collection("IS_TRAINING"))
        mean, variance = tf.python.control_flow_ops.cond(
            is_training, update_mean_var, lambda: (ema_mean, ema_var))
        return tf.nn.batch_norm_with_global_normalization(
            tensor_in, mean, variance, beta, gamma, epsilon,
            scale_after_normalization=scale_after_normalization)



#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import random

from sklearn import datasets
from sklearn.metrics import accuracy_score

import tensorflow as tf

from skflow.io import *
import skflow


class IOTest(tf.test.TestCase):

    def test_pandas_dataframe(self):
        if HAS_PANDAS:
            random.seed(42)
            iris = datasets.load_iris()
            data = pd.DataFrame(iris.data)
            labels = pd.DataFrame(iris.target)
            classifier = skflow.TensorFlowLinearClassifier(n_classes=3)
            classifier.fit(data, labels)
            score = accuracy_score(labels, classifier.predict(data))
            self.assertGreater(score, 0.5, "Failed with score = {0}".format(score))
        else:
            print("No pandas installed. pandas-related tests are skipped.")

    def test_pandas_series(self):
        if HAS_PANDAS:
            random.seed(42)
            iris = datasets.load_iris()
            data = pd.DataFrame(iris.data)
            labels = pd.Series(iris.target)
            classifier = skflow.TensorFlowLinearClassifier(n_classes=3)
            classifier.fit(data, labels)
            score = accuracy_score(labels, classifier.predict(data))
            self.assertGreater(score, 0.5, "Failed with score = {0}".format(score))

    def test_string_data_formats(self):
        with self.assertRaises(ValueError):
            skflow.io.extract_pandas_data(pd.DataFrame({"Test": ["A", "B"]}))
        with self.assertRaises(ValueError):
            skflow.io.extract_pandas_labels(pd.DataFrame({"Test": ["A", "B"]}))
   
    def test_dask_io(self):
        if HAS_DASK and HAS_PANDAS:
            # test dask.dataframe
            df = pd.DataFrame(dict(a=list('aabbcc'), b=list(range(6))),
                              index=pd.date_range(start='20100101', periods=6))
            ddf = dd.from_pandas(df, npartitions=3)
            extracted_ddf = extract_dask_data(ddf)
            self.assertEqual(extracted_ddf.divisions, (0, 2, 4, 6),
                             "Failed with divisions = {0}".format(extracted_ddf.divisions))
            self.assertEqual(extracted_ddf.columns, ('a', 'b'),
                             "Failed with columns = {0}".format(extracted_ddf.columns))
            # test dask.series
            labels = ddf['a']
            extracted_labels = extract_dask_labels(labels)
            self.assertEqual(extracted_labels.divisions, (0, 2, 4, 6),
                             "Failed with divisions = {0}".format(extracted_labels.divisions))
            # labels should only have one column
            with self.assertRaises(ValueError):
                extract_dask_labels(ddf)
        else:
            print("No dask installed. dask-related tests are skipped.")

    def test_dask_iris_classification(self):
        if HAS_DASK and HAS_PANDAS:
            random.seed(42)
            iris = datasets.load_iris()
            data = pd.DataFrame(iris.data)
            data = dd.from_pandas(data, npartitions=2)
            labels = pd.DataFrame(iris.target)
            labels = dd.from_pandas(labels, npartitions=2)
            classifier = skflow.TensorFlowLinearClassifier(n_classes=3)
            classifier.fit(data, labels)
            predictions = data.map_partitions(classifier.predict).compute()
            score = accuracy_score(labels.compute(), predictions)
            self.assertGreater(score, 0.5, "Failed with score = {0}".format(score))

if __name__ == '__main__':
    tf.test.main()



#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import random

from sklearn import datasets
from sklearn.metrics import accuracy_score, mean_squared_error, log_loss

import numpy as np
import tensorflow as tf

import skflow
from skflow.estimators import base


class BaseTest(tf.test.TestCase):

    def testOneDim(self):
        random.seed(42)
        X = np.random.rand(1000)
        y = 2 * X + 3
        regressor = skflow.TensorFlowLinearRegressor()
        regressor.fit(X, y)
        score = mean_squared_error(regressor.predict(X), y)
        self.assertLess(score, 0.3, "Failed with score = {0}".format(score))

    def testIris(self):
        iris = datasets.load_iris()
        classifier = skflow.TensorFlowLinearClassifier(n_classes=3)
        classifier.fit(iris.data, iris.target)
        score = accuracy_score(iris.target, classifier.predict(iris.data))
        self.assertGreater(score, 0.7, "Failed with score = {0}".format(score))

    def testIrisClassWeight(self):
        iris = datasets.load_iris()
        classifier = skflow.TensorFlowLinearClassifier(
            n_classes=3, class_weight=[0.1, 0.8, 0.1])
        classifier.fit(iris.data, iris.target)
        score = accuracy_score(iris.target, classifier.predict(iris.data))
        self.assertLess(score, 0.7, "Failed with score = {0}".format(score))

    def testIrisSummaries(self):
        iris = datasets.load_iris()
        classifier = skflow.TensorFlowLinearClassifier(n_classes=3)
        classifier.fit(iris.data, iris.target, logdir='/tmp/skflow_tests/')
        score = accuracy_score(iris.target, classifier.predict(iris.data))
        self.assertGreater(score, 0.5, "Failed with score = {0}".format(score))


    def testIrisContinueTraining(self):
        iris = datasets.load_iris()
        classifier = skflow.TensorFlowLinearClassifier(n_classes=3,
            learning_rate=0.01, continue_training=True, steps=250)
        classifier.fit(iris.data, iris.target)
        score1 = accuracy_score(iris.target, classifier.predict(iris.data))
        classifier.fit(iris.data, iris.target)
        score2 = accuracy_score(iris.target, classifier.predict(iris.data))
        self.assertGreater(score2, score1,
                           "Failed with score = {0}".format(score2))

    def testIrisStreaming(self):
        iris = datasets.load_iris()

        def iris_data():
            while True:
                for x in iris.data:
                    yield x

        def iris_predict_data():
            for x in iris.data:
                yield x

        def iris_target():
            while True:
                for y in iris.target:
                    yield y

        classifier = skflow.TensorFlowLinearClassifier(n_classes=3, steps=100)
        classifier.fit(iris_data(), iris_target())
        score1 = accuracy_score(iris.target, classifier.predict(iris.data))
        score2 = accuracy_score(iris.target, classifier.predict(iris_predict_data()))
        self.assertGreater(score1, 0.5, "Failed with score = {0}".format(score1))
        self.assertEqual(score2, score1, "Scores from {0} iterator doesn't "
                                         "match score {1} from full "
                                         "data.".format(score2, score1))

    def testIris_proba(self):
        random.seed(42)
        iris = datasets.load_iris()
        classifier = skflow.TensorFlowClassifier(n_classes=3, steps=250)
        classifier.fit(iris.data, iris.target)
        score = log_loss(iris.target, classifier.predict_proba(iris.data))
        self.assertLess(score, 0.8, "Failed with score = {0}".format(score))

    def testBoston(self):
        random.seed(42)
        boston = datasets.load_boston()
        regressor = skflow.TensorFlowLinearRegressor(
            batch_size=boston.data.shape[0],
            steps=500,
            learning_rate=0.001)
        regressor.fit(boston.data, boston.target)
        score = mean_squared_error(
            boston.target, regressor.predict(boston.data))
        self.assertLess(score, 150, "Failed with score = {0}".format(score))

    def testUnfitted(self):
        estimator = skflow.TensorFlowEstimator(model_fn=None, n_classes=1)
        with self.assertRaises(base.NotFittedError):
            estimator.predict([1, 2, 3])
        with self.assertRaises(base.NotFittedError):
            estimator.save('/tmp/path')


if __name__ == '__main__':
    tf.test.main()



#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import random

from sklearn import datasets
from sklearn.metrics import accuracy_score, mean_squared_error

import tensorflow as tf

import skflow

class NonLinearTest(tf.test.TestCase):

    def testIrisDNN(self):
        random.seed(42)
        iris = datasets.load_iris()
        classifier = skflow.TensorFlowDNNClassifier(
            hidden_units=[10, 20, 10], n_classes=3)
        classifier.fit(iris.data, iris.target)
        score = accuracy_score(iris.target, classifier.predict(iris.data))
        self.assertGreater(score, 0.5, "Failed with score = {0}".format(score))
        weights = classifier.weights_
        self.assertEqual(weights[0].shape, (4, 10))
        self.assertEqual(weights[1].shape, (10, 20))
        self.assertEqual(weights[2].shape, (20, 10))
        self.assertEqual(weights[3].shape, (10, 3))
        biases = classifier.bias_
        self.assertEqual(len(biases), 4)

    def testBostonDNN(self):
        random.seed(42)
        boston = datasets.load_boston()
        regressor = skflow.TensorFlowDNNRegressor(
            hidden_units=[10, 20, 10], n_classes=0,
            batch_size=boston.data.shape[0],
            steps=200, learning_rate=0.001)
        regressor.fit(boston.data, boston.target)
        score = mean_squared_error(
            boston.target, regressor.predict(boston.data))
        self.assertLess(score, 100, "Failed with score = {0}".format(score))
        weights = regressor.weights_
        self.assertEqual(weights[0].shape, (13, 10))
        self.assertEqual(weights[1].shape, (10, 20))
        self.assertEqual(weights[2].shape, (20, 10))
        self.assertEqual(weights[3].shape, (10, 1))
        biases = regressor.bias_
        self.assertEqual(len(biases), 4)

    def testRNN(self):
        random.seed(42)
        import numpy as np
        data = np.array(list([[2, 1, 2, 2, 3],
                              [2, 2, 3, 4, 5],
                              [3, 3, 1, 2, 1],
                              [2, 4, 5, 4, 1]]), dtype=np.float32)
        # labels for classification
        labels = np.array(list([1, 0, 1, 0]), dtype=np.float32)
        # targets for regression
        targets = np.array(list([10, 16, 10, 16]), dtype=np.float32)
        test_data = np.array(list([[1, 3, 3, 2, 1], [2, 3, 4, 5, 6]]))
        def input_fn(X):
            return tf.split(1, 5, X)

        # Classification
        classifier = skflow.TensorFlowRNNClassifier(
            rnn_size=2, cell_type='lstm', n_classes=2, input_op_fn=input_fn)
        classifier.fit(data, labels)
        classifier.weights_
        classifier.bias_
        predictions = classifier.predict(test_data)
        self.assertAllClose(predictions, np.array([1, 0]))
        
        classifier = skflow.TensorFlowRNNClassifier(
            rnn_size=2, cell_type='rnn', n_classes=2,
            input_op_fn=input_fn, num_layers=2)
        classifier.fit(data, labels)
        classifier = skflow.TensorFlowRNNClassifier(
            rnn_size=2, cell_type='invalid_cell_type', n_classes=2,
            input_op_fn=input_fn, num_layers=2)
        with self.assertRaises(ValueError):
            classifier.fit(data, labels)

        # Regression
        regressor = skflow.TensorFlowRNNRegressor(
            rnn_size=2, cell_type='gru', input_op_fn=input_fn)
        regressor.fit(data, targets)
        regressor.weights_
        regressor.bias_
        predictions = regressor.predict(test_data)
    
    def testBidirectionalRNN(self):
        random.seed(42)
        import numpy as np
        data = np.array(list([[2, 1, 2, 2, 3],
                              [2, 2, 3, 4, 5],
                              [3, 3, 1, 2, 1],
                              [2, 4, 5, 4, 1]]), dtype=np.float32)
        labels = np.array(list([1, 0, 1, 0]), dtype=np.float32)
        def input_fn(X):
            return tf.split(1, 5, X)

        # Classification
        classifier = skflow.TensorFlowRNNClassifier(
            rnn_size=2, cell_type='lstm', n_classes=2, input_op_fn=input_fn,
            bidirectional=True)
        classifier.fit(data, labels)
        predictions = classifier.predict(np.array(list([[1, 3, 3, 2, 1],
                                                        [2, 3, 4, 5, 6]])))
        self.assertAllClose(predictions, np.array([1, 0]))
        

if __name__ == "__main__":
    tf.test.main()



"""Base estimator class."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function, absolute_import

import json
import os
import datetime
from six import string_types

import numpy as np
import tensorflow as tf

from google.protobuf import text_format

from sklearn.base import BaseEstimator
try:
    from sklearn.exceptions import NotFittedError
except ImportError:
    from sklearn.utils.validation import NotFittedError  # pylint: disable=ungrouped-imports

from skflow.trainer import TensorFlowTrainer, RestoredTrainer
from skflow.io.data_feeder import setup_train_data_feeder
from skflow.io.data_feeder import setup_predict_data_feeder
from skflow.ops.dropout_ops import DROPOUTS


class TensorFlowEstimator(BaseEstimator):
    """Base class for all TensorFlow estimators.

    Parameters:
        model_fn: Model function, that takes input X, y tensors and outputs
                  prediction and loss tensors.
        n_classes: Number of classes in the target.
        tf_master: TensorFlow master. Empty string is default for local.
        batch_size: Mini batch size.
        steps: Number of steps to run over data.
        optimizer: Optimizer name (or class), for example "SGD", "Adam",
                   "Adagrad".
        learning_rate: If this is constant float value, no decay function is used.
            Instead, a customized decay function can be passed that accepts
            global_step as parameter and returns a Tensor.
            e.g. exponential decay function:
            def exp_decay(global_step):
                return tf.train.exponential_decay(
                    learning_rate=0.1, global_step,
                    decay_steps=2, decay_rate=0.001)
        class_weight: None or list of n_classes floats. Weight associated with
                     classes for loss computation. If not given, all classes are suppose to have
                     weight one.
        tf_random_seed: Random seed for TensorFlow initializers.
            Setting this value, allows consistency between reruns.
        continue_training: when continue_training is True, once initialized
            model will be continuely trained on every call of fit.
        num_cores: Number of cores to be used. (default: 4)
        verbose: Controls the verbosity, possible values:
                 0: the algorithm and debug information is muted.
                 1: trainer prints the progress.
                 2: log device placement is printed.
        early_stopping_rounds: Activates early stopping if this is not None.
            Loss needs to decrease at least every every <early_stopping_rounds>
            round(s) to continue training. (default: None)
        max_to_keep: The maximum number of recent checkpoint files to keep.
            As new files are created, older files are deleted.
            If None or 0, all checkpoint files are kept.
            Defaults to 5 (that is, the 5 most recent checkpoint files are kept.)
        keep_checkpoint_every_n_hours: Number of hours between each checkpoint
            to be saved. The default value of 10,000 hours effectively disables the feature.
    """

    def __init__(self, model_fn, n_classes, tf_master="", batch_size=32,
                 steps=200, optimizer="SGD",
                 learning_rate=0.1, class_weight=None,
                 tf_random_seed=42, continue_training=False,
                 num_cores=4, verbose=1, early_stopping_rounds=None,
                 max_to_keep=5, keep_checkpoint_every_n_hours=10000):
        self.n_classes = n_classes
        self.tf_master = tf_master
        self.batch_size = batch_size
        self.steps = steps
        self.verbose = verbose
        self.optimizer = optimizer
        self.learning_rate = learning_rate
        self.tf_random_seed = tf_random_seed
        self.model_fn = model_fn
        self.continue_training = continue_training
        self.num_cores = num_cores
        self._initialized = False
        self._early_stopping_rounds = early_stopping_rounds
        self.max_to_keep = max_to_keep
        self.keep_checkpoint_every_n_hours = keep_checkpoint_every_n_hours
        self.class_weight = class_weight

    def _setup_training(self):
        """Sets up graph, model and trainer."""
        self._graph = tf.Graph()
        self._graph.add_to_collection("IS_TRAINING", True)
        with self._graph.as_default():
            tf.set_random_seed(self.tf_random_seed)
            self._global_step = tf.Variable(
                0, name="global_step", trainable=False)

            # Setting up input and output placeholders.
            input_shape = [None] + self._data_feeder.input_shape[1:]
            output_shape = [None] + self._data_feeder.output_shape[1:]
            self._inp = tf.placeholder(
                tf.as_dtype(self._data_feeder.input_dtype), input_shape,
                name="input")
            self._out = tf.placeholder(
                tf.as_dtype(self._data_feeder.output_dtype), output_shape,
                name="output")

            # If class weights are provided, add them to the graph.
            # Different loss functions can use this tensor by name.
            if self.class_weight:
                self._class_weight_node = tf.constant(
                    self.class_weight, name='class_weight')

            # Add histograms for X and y if they are floats.
            if self._data_feeder.input_dtype in (np.float32, np.float64):
                tf.histogram_summary("X", self._inp)
            if self._data_feeder.output_dtype in (np.float32, np.float64):
                tf.histogram_summary("y", self._out)

            # Create model's graph.
            self._model_predictions, self._model_loss = self.model_fn(
                self._inp, self._out)

            # Create summary to monitor loss
            tf.scalar_summary("loss", self._model_loss)

            # Set up a single operator to merge all the summaries
            self._summaries = tf.merge_all_summaries()

            # Create trainer and augment graph with gradients and optimizer.
            # Additionally creates initialization ops.
            self._trainer = TensorFlowTrainer(
                loss=self._model_loss, global_step=self._global_step,
                optimizer=self.optimizer, learning_rate=self.learning_rate)

            # Create model's saver capturing all the nodes created up until now.
            self._saver = tf.train.Saver(
                max_to_keep=self.max_to_keep,
                keep_checkpoint_every_n_hours=self.keep_checkpoint_every_n_hours)

            # Create session to run model with.
            self._session = tf.Session(self.tf_master,
                                       config=tf.ConfigProto(
                                           log_device_placement=self.verbose > 1,
                                           inter_op_parallelism_threads=self.num_cores,
                                           intra_op_parallelism_threads=self.num_cores))

    def _setup_summary_writer(self, logdir):
        """Sets up the summary writer to prepare for later optional visualization."""
        self._summary_writer = tf.train.SummaryWriter(
            os.path.join(logdir, datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')),
            graph_def=self._session.graph_def)

    def fit(self, X, y, logdir=None):
        """Builds a neural network model given provided `model_fn` and training
        data X and y.

        Note: called first time constructs the graph and initializers
        variables. Consecutives times it will continue training the same model.
        This logic follows partial_fit() interface in scikit-learn.

        To restart learning, create new estimator.

        Args:
            X: matrix or tensor of shape [n_samples, n_features...]. Can be
            iterator that returns arrays of features. The training input
            samples for fitting the model.
            y: vector or matrix [n_samples] or [n_samples, n_outputs]. Can be
            iterator that returns array of targets. The training target values
            (class labels in classification, real numbers in regression).
            logdir: the directory to save the log file that can be used for
            optional visualization.

        Returns:
            Returns self.
        """
        # Sets up data feeder.
        self._data_feeder = setup_train_data_feeder(X, y,
                                                    self.n_classes,
                                                    self.batch_size)
        if not self.continue_training or not self._initialized:
            # Sets up model and trainer.
            self._setup_training()
            # Initialize model parameters.
            self._trainer.initialize(self._session)
            self._initialized = True

        # Sets up summary writer for later optional visualization.
        # Due to not able to setup _summary_writer in __init__ as it's not a
        # parameter of the model, here we need to check if such variable exists
        # and if it's None or not (in case it was setup in a previous run).
        # It is initialized only in the case where it wasn't before and log dir
        # is provided.
        if logdir:
            if (not hasattr(self, "_summary_writer") or
                    (hasattr(self, "_summary_writer") and self._summary_writer is None)):
                self._setup_summary_writer(logdir)
        else:
            self._summary_writer = None

        # Train model for given number of steps.
        self._trainer.train(self._session,
                            self._data_feeder.get_feed_dict_fn(
                                self._inp, self._out),
                            self.steps,
                            self._summary_writer,
                            self._summaries,
                            verbose=self.verbose,
                            early_stopping_rounds=self._early_stopping_rounds,
                            feed_params_fn=self._data_feeder.get_feed_params)
        return self

    def partial_fit(self, X, y):
        """Incremental fit on a batch of samples.

        This method is expected to be called several times consecutively
        on different or the same chunks of the dataset. This either can
        implement iterative training or out-of-core/online training.

        This is especially useful when the whole dataset is too big to
        fit in memory at the same time. Or when model is taking long time
        to converge, and you want to split up training into subparts.

        Args:
            X: matrix or tensor of shape [n_samples, n_features...]. Can be
            iterator that returns arrays of features. The training input
            samples for fitting the model.
            y: vector or matrix [n_samples] or [n_samples, n_outputs]. Can be
            iterator that returns array of targets. The training target values
            (class label in classification, real numbers in regression).

        Returns:
            Returns self.
        """
        return self.fit(X, y)

    def _predict(self, X, batch_size=-1):
        if not self._initialized:
            raise NotFittedError()
        self._graph.add_to_collection("IS_TRAINING", False)
        predict_data_feeder = setup_predict_data_feeder(
            X, batch_size=batch_size)
        preds = []
        dropouts = self._graph.get_collection(DROPOUTS)
        feed_dict = {prob: 1.0 for prob in dropouts}
        for data in predict_data_feeder:
            feed_dict[self._inp] = data
            preds.append(self._session.run(
                self._model_predictions,
                feed_dict))
        return np.concatenate(preds, axis=0)

    def predict(self, X, axis=1, batch_size=-1):
        """Predict class or regression for X.

        For a classification model, the predicted class for each sample in X is
        returned. For a regression model, the predicted value based on X is
        returned.

        Args:
            X: array-like matrix, [n_samples, n_features...] or iterator.
            axis: Which axis to argmax for classification.
                  By default axis 1 (next after batch) is used.
                  Use 2 for sequence predictions.
            batch_size: If test set is too big, use batch size to split
                        it into mini batches. By default full dataset is used.

        Returns:
            y: array of shape [n_samples]. The predicted classes or predicted
            value.
        """
        pred = self._predict(X, batch_size=batch_size)
        if self.n_classes < 2:
            return pred
        return pred.argmax(axis=axis)

    def predict_proba(self, X, batch_size=-1):
        """Predict class probability of the input samples X.

        Args:
            X: array-like matrix, [n_samples, n_features...] or iterator.
            batch_size: If test set is too big, use batch size to split
                        it into mini batches. By default full dataset is used.

        Returns:
            y: array of shape [n_samples, n_classes]. The predicted
            probabilities for each class.

       """
        return self._predict(X, batch_size=batch_size)

    def get_tensor(self, name):
        """Returns tensor by name.

        Args:
            name: string, name of the tensor.

        Returns:
            Tensor.
        """
        return self._graph.get_tensor_by_name(name)

    def get_tensor_value(self, name):
        """Returns value of the tensor give by name.

        Args:
            name: string, name of the tensor.

        Returns:
            Numpy array - value of the tensor.
        """
        return self._session.run(self.get_tensor(name))

    def save(self, path):
        """Saves checkpoints and graph to given path.

        Args:
            path: Folder to save model to.
        """
        if not self._initialized:
            raise NotFittedError()

        # Currently Saver requires absolute path to work correctly.
        path = os.path.abspath(path)

        if not os.path.exists(path):
            os.makedirs(path)
        if not os.path.isdir(path):
            raise ValueError("Path %s should be a directory to save"
                             "checkpoints and graph." % path)
        with open(os.path.join(path, 'model.def'), 'w') as fmodel:
            all_params = self.get_params()
            params = {}
            for key, value in all_params.items():
                if not callable(value) and value is not None:
                    params[key] = value
            params['class_name'] = type(self).__name__
            fmodel.write(json.dumps(params))
        with open(os.path.join(path, 'endpoints'), 'w') as foutputs:
            foutputs.write('%s\n%s\n%s\n%s' % (
                self._inp.name,
                self._out.name,
                self._model_predictions.name,
                self._model_loss.name))
        with open(os.path.join(path, 'graph.pbtxt'), 'w') as fgraph:
            fgraph.write(str(self._graph.as_graph_def()))
        with open(os.path.join(path, 'saver.pbtxt'), 'w') as fsaver:
            fsaver.write(str(self._saver.as_saver_def()))
        self._saver.save(self._session, os.path.join(path, 'model'),
                         global_step=self._global_step)

    def _restore(self, path):
        """Restores this estimator from given path.

        Note: will rebuild the graph and initialize all parameters,
        and will ignore provided model.

        Args:
            path: Path to checkpoints and other information.
        """
        # Currently Saver requires absolute path to work correctly.
        path = os.path.abspath(path)

        self._graph = tf.Graph()
        with self._graph.as_default():
            endpoints_filename = os.path.join(path, 'endpoints')
            if not os.path.exists(endpoints_filename):
                raise ValueError("Restore folder doesn't contain endpoints.")
            with open(endpoints_filename) as foutputs:
                endpoints = foutputs.read().split('\n')
            graph_filename = os.path.join(path, 'graph.pbtxt')
            if not os.path.exists(graph_filename):
                raise ValueError("Restore folder doesn't contain graph definition.")
            with open(graph_filename) as fgraph:
                graph_def = tf.GraphDef()
                text_format.Merge(fgraph.read(), graph_def)
                (self._inp, self._out,
                 self._model_predictions, self._model_loss) = tf.import_graph_def(
                     graph_def, name='', return_elements=endpoints)
            saver_filename = os.path.join(path, 'saver.pbtxt')
            if not os.path.exists(saver_filename):
                raise ValueError("Restore folder doesn't contain saver defintion.")
            with open(saver_filename) as fsaver:
                saver_def = tf.python.training.saver_pb2.SaverDef()
                text_format.Merge(fsaver.read(), saver_def)
                self._saver = tf.train.Saver(saver_def=saver_def)

            # Restore trainer
            self._global_step = self._graph.get_tensor_by_name('global_step:0')
            trainer_op = self._graph.get_operation_by_name('train')
            self._trainer = RestoredTrainer(
                self._model_loss, self._global_step, trainer_op)

            # Restore summaries.
            self._summaries = self._graph.get_operation_by_name('MergeSummary/MergeSummary')

            # Restore session.
            self._session = tf.Session(self.tf_master,
                                       config=tf.ConfigProto(
                                           log_device_placement=self.verbose > 1,
                                           inter_op_parallelism_threads=self.num_cores,
                                           intra_op_parallelism_threads=self.num_cores))
            checkpoint_path = tf.train.latest_checkpoint(path)
            if checkpoint_path is None:
                raise ValueError("Missing checkpoint files in the %s. Please "
                                 "make sure you are you have checkpoint file that describes "
                                 "latest checkpoints and appropriate checkpoints are there. "
                                 "If you have moved the folder, you at this point need to "
                                 "update manually update the paths in the checkpoint file." % path)
            self._saver.restore(self._session, checkpoint_path)
        # Set to be initialized.
        self._initialized = True

    @classmethod
    def restore(cls, path):
        """Restores model from give path.

        Args:
            path: Path to the checkpoints and other model information.

        Returns:
            Estiamator, object of the subclass of TensorFlowEstimator.
        """
        model_def_filename = os.path.join(path, 'model.def')
        if not os.path.exists(model_def_filename):
            raise ValueError("Restore folder doesn't contain model definition.")
        with open(model_def_filename) as fmodel:
            model_def = json.loads(fmodel.read())
            # TensorFlow binding requires parameters to be strings not unicode.
            # Only issue in Python2.
            for key, value in model_def.items():
                if (isinstance(value, string_types) and
                        not isinstance(value, str)):
                    model_def[key] = str(value)
        class_name = model_def.pop('class_name')
        if class_name == 'TensorFlowEstimator':
            custom_estimator = TensorFlowEstimator(model_fn=None, **model_def)
            custom_estimator._restore(path)
            return custom_estimator

        # To avoid cyclical dependencies, import inside the function instead of
        # the beginning of the file.
        from skflow import estimators
        # Estimator must be one of the defined estimators in the __init__ file.
        estimator = getattr(estimators, class_name)(**model_def)
        estimator._restore(path)
        return estimator




"""Implements preprocesing transformers for categorical variables."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import math
import numpy as np

from skflow.preprocessing import categorical_vocabulary
from skflow.io.data_feeder import setup_processor_data_feeder


class CategoricalProcessor(object):
    """Maps documents to sequences of word ids.

    As a common convention, Nan values are handled as unknown tokens.
    Both float('nan') and np.nan are accepted.

    Parameters:
        min_frequency: Minimum frequency of categories in the vocabulary.
        share: Share vocabulary between variables.
        vocabularies: list of CategoricalVocabulary objects for each variable in
                    the input dataset.

    Attributes:
        vocabularies_: CategoricalVocabulary object.
    """

    def __init__(self, min_frequency=0, share=False, vocabularies=None):
        self.min_frequency = min_frequency
        self.share = share
        self.vocabularies_ = vocabularies

    def freeze(self, freeze=True):
        """Freeze or unfreeze all vocabularies.

        Args:
            freeze: Boolean, indicate if vocabularies should be frozen.
        """
        for vocab in self.vocabularies_:
            vocab.freeze(freeze)

    def fit(self, X, unused_y=None):
        """Learn a vocabulary dictionary of all categories in X.

        Args:
            raw_documents: numpy matrix or iterable of lists/numpy arrays.
            unused_y: to match fit format signature of estimators.

        Returns:
            self
        """
        X = setup_processor_data_feeder(X)
        for row in X:
            # Create vocabularies if not given.
            if self.vocabularies_ is None:
                # If not share, one per column, else one shared across.
                if not self.share:
                    self.vocabularies_ = [
                        categorical_vocabulary.CategoricalVocabulary() for _ in row]
                else:
                    vocab = categorical_vocabulary.CategoricalVocabulary()
                    self.vocabularies_ = [vocab for _ in row]
            for idx, value in enumerate(row):
                # Nans are handled as unknowns.
                if (isinstance(value, float) and math.isnan(value)) or value == np.nan:
                    continue
                self.vocabularies_[idx].add(value)
        if self.min_frequency > 0:
            for vocab in self.vocabularies_:
                vocab.trim(self.min_frequency)
        self.freeze()
        return self

    def fit_transform(self, X, unused_y=None):
        """Learn the vocabulary dictionary and return indexies of categories.

        Args:
            X: numpy matrix or iterable of lists/numpy arrays.
            unused_y: to match fit_transform signature of estimators.

        Returns:
            X: iterable, [n_samples]. Category-id matrix.
        """
        self.fit(X)
        return self.transform(X)

    def transform(self, X):
        """Transform documents to category-id matrix.

        Converts categories to ids give fitted vocabulary from `fit` or
        one provided in the constructor.

        Args:
            X: numpy matrix or iterable of lists/numpy arrays.

        Returns:
            X: iterable, [n_samples]. Category-id matrix.
        """
        self.freeze()
        X = setup_processor_data_feeder(X)
        for row in X:
            output_row = []
            for idx, value in enumerate(row):
                # Return <UNK> when it's Nan.
                if (isinstance(value, float) and math.isnan(value)) or value == np.nan:
                    output_row.append(0)
                    continue
                output_row.append(self.vocabularies_[idx].get(value))
            yield np.array(output_row, dtype=np.int64)




#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import numpy as np
import tensorflow as tf

from skflow import ops


class OpsTest(tf.test.TestCase):

    def test_softmax_classifier(self):
        with self.test_session() as session:
            features = tf.placeholder(tf.float32, [None, 3])
            labels = tf.placeholder(tf.float32, [None, 2])
            weights = tf.constant([[0.1, 0.1], [0.1, 0.1], [0.1, 0.1]])
            biases = tf.constant([0.2, 0.3])
            class_weight = tf.constant([0.1, 0.9])
            prediction, loss = ops.softmax_classifier(features, labels, weights, biases, class_weight)
            self.assertEqual(prediction.get_shape()[1], 2)
            self.assertEqual(loss.get_shape(), [])
            value = session.run(loss, {features: [[0.2, 0.3, 0.2]], labels: [[0, 1]]})
            self.assertAllClose(value, 0.55180627)
 
    def test_embedding_lookup(self):
        d_embed = 5
        n_embed = 10
        ids_shape = (2, 3, 4)
        embeds = np.random.randn(n_embed, d_embed)
        ids = np.random.randint(0, n_embed, ids_shape)
        with self.test_session():
            embed_np = embeds[ids]
            embed_tf = ops.embedding_lookup(embeds, ids).eval()
        self.assertEqual(embed_np.shape, embed_tf.shape)
        self.assertAllClose(embed_np, embed_tf)

    def test_categorical_variable(self):
        tf.set_random_seed(42)
        with self.test_session() as sess:
            cat_var_idx = tf.placeholder(tf.int64, [2, 2])
            embeddings = ops.categorical_variable(cat_var_idx, n_classes=5,
                                                  embedding_size=10,
                                                  name="my_cat_var")
            sess.run(tf.initialize_all_variables())
            emb1 = sess.run(embeddings, feed_dict={cat_var_idx.name: [[0, 1],
                                                                      [2, 3]]})
            emb2 = sess.run(embeddings, feed_dict={cat_var_idx.name: [[0, 2],
                                                                      [1, 3]]})
        self.assertEqual(emb1.shape, emb2.shape)
        self.assertAllEqual(np.transpose(emb2, axes=[1, 0, 2]), emb1)

    def test_conv2d(self):
        tf.set_random_seed(42)
        batch_size = 32
        input_shape = (10, 10)
        n_filters = 7
        filter_shape = (5, 5)
        vals = np.random.randn(batch_size, input_shape[0], input_shape[1], 1)
        with self.test_session() as sess:
            tf.add_to_collection("IS_TRAINING", True)
            tensor_in = tf.placeholder(tf.float32, [batch_size, input_shape[0],
                                                    input_shape[1], 1])
            res = ops.conv2d(
                tensor_in, n_filters, filter_shape, batch_norm=True)
            sess.run(tf.initialize_all_variables())
            conv = sess.run(res, feed_dict={tensor_in.name: vals})
        self.assertEqual(conv.shape, (batch_size, input_shape[0],
                                      input_shape[1], n_filters))

    def test_one_hot_matrix(self):
        with self.test_session() as sess:
            tensor_in = tf.placeholder(tf.int64, [10, 2])
            one_hot_tensor = ops.one_hot_matrix(tensor_in, 3)
            res = sess.run(ops.one_hot_matrix([[0, 1], [2, 1]], 3))
        self.assertAllEqual(one_hot_tensor.get_shape(), [10, 2, 3])
        self.assertAllEqual(res, [[[1.0, 0, 0], [0, 1.0, 0]], 
                                  [[0, 0, 1.0], [0, 1.0, 0]]])


if __name__ == '__main__':
    tf.test.main()



"""Methods to allow dask.DataFrame."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import numpy as np


try:
    import dask.dataframe as dd
    allowed_classes = (dd.Series, dd.DataFrame)
    HAS_DASK = True
except ImportError:
    HAS_DASK = False


def _add_to_index(df, start):
    """Make a new dask.dataframe where we add these values to the
    index of each subdataframe.
    """
    df = df.copy()
    df.index = df.index + start
    return df


def _get_divisions(df):
    """Number of rows in each sub-dataframe"""
    lengths = df.map_partitions(len).compute()
    divisions = np.cumsum(lengths).tolist()
    divisions.insert(0, 0)
    return divisions


def _construct_dask_df_with_divisions(df):
    """Construct the new task graph and make a new dask.dataframe around it"""
    divisions = _get_divisions(df)
    name = 'csv-index' + df._name
    dsk = {(name, i): (_add_to_index, (df._name, i), divisions[i]) for i in range(df.npartitions)}
    from toolz import merge
    if isinstance(df, dd.DataFrame):
        return dd.DataFrame(merge(dsk, df.dask), name, df.columns, divisions)
    elif isinstance(df, dd.Series):
        return dd.Series(merge(dsk, df.dask), name, df.name, divisions)


def extract_dask_data(data):
    """Extract data from dask.Series or dask.DataFrame for predictors"""
    if isinstance(data, allowed_classes):
        return _construct_dask_df_with_divisions(data)
    else:
        return data


def extract_dask_labels(labels):
    """Extract data from dask.Series for labels"""
    if isinstance(labels, allowed_classes):
        if len(labels.columns) > 1:
            raise ValueError('Only one column for labels is allowed.')
        return _construct_dask_df_with_divisions(labels)
    else:
        return labels



"""TensorFlow ops for Convolution NNs."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function, absolute_import

import tensorflow as tf
from skflow.ops.batch_norm_ops import batch_normalize


def conv2d(tensor_in, n_filters, filter_shape, strides=None, padding='SAME',
           bias=True, activation=None, batch_norm=False):
    """Creates 2D convolutional subgraph with bank of filters.

    Uses tf.nn.conv2d under the hood.
    Creates a filter bank:
      [filter_shape[0], filter_shape[1], tensor_in[3], n_filters]
    and applies it to the input tensor.

    Args:
        tensor_in: input Tensor, 4D shape:
                   [batch, in_height, in_width, in_depth].
        n_filters: number of filters in the bank.
        filter_shape: Shape of filters, a list of ints, 1-D of length 2.
        strides: A list of ints, 1-D of length 4. The stride of the sliding
                 window for each dimension of input.
        padding: A string: 'SAME' or 'VALID'. The type of padding algorthim to
                 use.
        bias: Boolean, if to add bias.
        activation: Activation Op, optional. If provided applied on the output.
        batch_norm: Whether to apply batch normalization.

    Returns:
        A Tensor with resulting convolution.
    """
    with tf.variable_scope('convolution'):
        if strides is None:
            strides = [1, 1, 1, 1]
        input_shape = tensor_in.get_shape()
        filter_shape = list(filter_shape) + [input_shape[3], n_filters]
        filters = tf.get_variable('filters', filter_shape, tf.float32)
        output = tf.nn.conv2d(tensor_in, filters, strides, padding)
        if bias:
            bias_var = tf.get_variable('bias', [1, 1, 1, n_filters],
                                       tf.float32)
            output = output + bias_var
        if batch_norm:
            output = batch_normalize(output)
        if activation:
            output = activation(output)
        return output




"""Implementations of different data feeders to provide data for TF trainer."""

#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function, absolute_import

import itertools
import math

import six
from six.moves import xrange   # pylint: disable=redefined-builtin

import numpy as np
from sklearn.utils import check_array

from skflow.io import HAS_PANDAS, extract_pandas_data, extract_pandas_matrix, extract_pandas_labels
from skflow.io import HAS_DASK, extract_dask_data, extract_dask_labels


def _get_in_out_shape(x_shape, y_shape, n_classes, batch_size):
    """Returns shape for input and output of the data feeder."""
    x_shape = list(x_shape[1:]) if len(x_shape) > 1 else [1]
    input_shape = [batch_size] + x_shape
    y_shape = list(y_shape[1:]) if len(y_shape) > 1 else []
    # Skip first dimention if it is 1.
    if y_shape and y_shape[0] == 1:
        y_shape = y_shape[1:]
    if n_classes > 1:
        output_shape = [batch_size] + y_shape + [n_classes]
    else:
        output_shape = [batch_size] + y_shape
    return input_shape, output_shape


def _data_type_filter(X, y):
    """Filter data types into acceptable format"""
    if HAS_DASK:
        X = extract_dask_data(X)
        y = extract_dask_labels(y)
    if HAS_PANDAS:
        X = extract_pandas_data(X)
        y = extract_pandas_labels(y)
    return X, y


def _is_iterable(X):
    return hasattr(X, 'next') or hasattr(X, '__next__')


def setup_train_data_feeder(X, y, n_classes, batch_size):
    """Create data feeder, to sample inputs from dataset.
    If X and y are iterators, use StreamingDataFeeder.

    Args:
        X: numpy, pandas or Dask matrix or iterable.
        y: numpy, pandas or Dask array or iterable.
        n_classes: number of classes.
        batch_size: size to split data into parts.

    Returns:
        DataFeeder object that returns training data.
    """
    X, y = _data_type_filter(X, y)
    if HAS_DASK:
        import dask.dataframe as dd
        allowed_classes = (dd.Series, dd.DataFrame)
        if isinstance(X, allowed_classes) and isinstance(y, allowed_classes):
            data_feeder_cls = DaskDataFeeder
        else:
            data_feeder_cls = DataFeeder
    else:
        data_feeder_cls = DataFeeder

    if _is_iterable(X):
        if not _is_iterable(y):
            raise ValueError("Both X and y should be iterators for "
                             "streaming learning to work.")
        data_feeder_cls = StreamingDataFeeder
    return data_feeder_cls(X, y, n_classes, batch_size)


def _batch_data(X, batch_size):
    chunk = []
    for data in X:
        chunk.append(data)
        if batch_size > 0 and len(chunk) > batch_size:
            yield np.matrix(chunk)
    yield np.matrix(chunk)


def setup_predict_data_feeder(X, batch_size=-1):
    """Returns an iterable for feeding into predict step.

    Args:
        X: numpy, pandas, Dask array or iterable.
        batch_size: Size of batches to split data into.
            If negative, returns one batch of full size.

    Returns:
        List or iterator of parts of data to predict on.
    """
    if HAS_DASK:
        X = extract_dask_data(X)
    if HAS_PANDAS:
        X = extract_pandas_data(X)
    if _is_iterable(X):
        return _batch_data(X, batch_size)
    if len(X.shape) == 1:
        X = np.reshape(X, (-1, 1))
    if batch_size > 0:
        n_batches = int(math.ceil(float(len(X)) / batch_size))
        return [X[i * batch_size:(i + 1) * batch_size] for i in xrange(n_batches)]
    return [X]


def setup_processor_data_feeder(X):
    """Sets up processor iterable.

    Args:
        X: numpy, pandas or iterable.

    Returns:
        Iterable of data to process.
    """
    if HAS_PANDAS:
        X = extract_pandas_matrix(X)
    return X


class DataFeeder(object):
    """Data feeder is an example class to sample data for TF trainer.

    Parameters:
        X: feature Nd numpy matrix of shape [n_samples, n_features, ...].
        y: target vector, either floats for regression or class id for
            classification. If matrix, will consider as a sequence
            of targets.
        n_classes: number of classes, 0 and 1 are considered regression.
        batch_size: mini batch size to accumulate.
        random_state: numpy RandomState object to reproduce sampling.

    Attributes:
        X: input features.
        y: input target.
        n_classes: number of classes.
        batch_size: mini batch size to accumulate.
        input_shape: shape of the input.
        output_shape: shape of the output.
        input_dtype: dtype of input.
        output_dtype: dtype of output.
    """

    def __init__(self, X, y, n_classes, batch_size, random_state=None):
        x_dtype = np.int64 if X.dtype == np.int64 else np.float32
        self.X = check_array(X, ensure_2d=False,
                             allow_nd=True, dtype=x_dtype)
        self.y = check_array(y, ensure_2d=False, dtype=np.float32)
        self.n_classes = n_classes
        self.batch_size = batch_size
        self.input_shape, self.output_shape = _get_in_out_shape(
            self.X.shape, self.y.shape, n_classes, batch_size)
        self.input_dtype, self.output_dtype = self.X.dtype, self.y.dtype
        self.random_state = np.random.RandomState(42) if random_state is None else random_state
        self.indices = self.random_state.permutation(self.X.shape[0])
        self.offset = 0
        self.epoch = 0

    def get_feed_params(self):
        """Function returns a dict with data feed params while training.
        Returns:
            A dict with data feed params while training.
        """
        return {
            'epoch': self.epoch,
            'offset': self.offset,
            'batch_size': self.batch_size
        }

    def get_feed_dict_fn(self, input_placeholder, output_placeholder):
        """Returns a function, that will sample data and provide it to given
        placeholders.

        Args:
            input_placeholder: tf.Placeholder for input features mini batch.
            output_placeholder: tf.Placeholder for output targets.
        Returns:
            A function that when called samples a random subset of batch size
            from X and y.
        """
        def _feed_dict_fn():
            # take random indices
            batch_indices = self.indices[self.offset: self.offset+self.batch_size]

            # assign input features from random indices
            inp = np.array(self.X[batch_indices]).reshape((batch_indices.shape[0], 1)) \
                if len(self.X.shape) == 1 else self.X[batch_indices]

            # assign labels from random indices
            self.output_shape[0] = batch_indices.shape[0]
            out = np.zeros(self.output_shape, dtype=self.output_dtype)
            for i in xrange(out.shape[0]):
                sample = batch_indices[i]
                if self.n_classes > 1:
                    if len(self.output_shape) == 2:
                        out.itemset((i, self.y[sample]), 1.0)
                    else:
                        for idx, value in enumerate(self.y[sample]):
                            out.itemset(tuple([i, idx, value]), 1.0)
                else:
                    out[i] = self.y[sample]

            # move offset and reset it if necessary
            self.offset += self.batch_size
            if self.offset >= self.X.shape[0]:
                self.indices = self.random_state.permutation(self.X.shape[0])
                self.offset = 0
                self.epoch += 1

            return {input_placeholder.name: inp, output_placeholder.name: out}
        return _feed_dict_fn


class StreamingDataFeeder(object):
    """Data feeder for TF trainer that reads data from iterator.

    Streaming data feeder allows to read data as it comes it from disk or
    somewhere else. It's custom to have this iterators rotate infinetly over
    the dataset, to allow control of how much to learn on the trainer side.

    Parameters:
        X: iterator that returns for each element, returns features.
        y: iterator that returns for each element, returns 1 or many classes /
           regression values.
        n_classes: indicator of how many classes the target has.
        batch_size: Mini batch size to accumulate.

    Attributes:
        X: input features.
        y: input target.
        n_classes: number of classes.
        batch_size: mini batch size to accumulate.
        input_shape: shape of the input.
        output_shape: shape of the output.
        input_dtype: dtype of input.
        output_dtype: dtype of output.
    """

    def __init__(self, X, y, n_classes, batch_size):
        X_first_el = six.next(X)
        y_first_el = six.next(y)
        self.X = itertools.chain([X_first_el], X)
        self.y = itertools.chain([y_first_el], y)
        self.n_classes = n_classes
        self.batch_size = batch_size
        self.input_shape, self.output_shape = _get_in_out_shape(
            [1] + list(X_first_el.shape),
            [1] + list(y_first_el.shape), n_classes, batch_size)
        self.input_dtype = X_first_el.dtype
        # Convert float64 to float32, as all the parameters in the model are
        # floats32 and there is a lot of benefits in using it in NNs.
        if self.input_dtype == np.float64:
            self.input_dtype = np.float32
        # Output types are floats, due to both softmaxes and regression req.
        self.output_dtype = np.float32

    def get_feed_params(self):
        """Function returns a dict with data feed params while training.
        Returns:
            A dict with data feed params while training.
        """
        return {'batch_size': self.batch_size}

    def get_feed_dict_fn(self, input_placeholder, output_placeholder):
        """Returns a function, that will sample data and provide it to given

        placeholders.

        Args:
            input_placeholder: tf.Placeholder for input features mini batch.
            output_placeholder: tf.Placeholder for output targets.
        Returns:
            A function that when called samples a random subset of batch size
            from X and y.
        """
        def _feed_dict_fn():
            inp = np.zeros(self.input_shape, dtype=self.input_dtype)
            out = np.zeros(self.output_shape, dtype=self.output_dtype)
            for i in xrange(self.batch_size):
                inp[i, :] = six.next(self.X)
                y = six.next(self.y)
                if self.n_classes > 1:
                    if len(self.output_shape) == 2:
                        out.itemset((i, y), 1.0)
                    else:
                        for idx, value in enumerate(y):
                            out.itemset(tuple([i, idx, value]), 1.0)
                else:
                    out[i] = y
            return {input_placeholder.name: inp, output_placeholder.name: out}
        return _feed_dict_fn


class DaskDataFeeder(object):
    """Data feeder for TF trainer that reads data from dask.Series and dask.DataFrame.

    Numpy arrays can be serialized to disk and it's possible to do random seeks into them.
    DaskDataFeeder will remove requirement to have full dataset in the memory and still do
    random seeks for sampling of batches.

    Parameters:
        X: iterator that returns for each element, returns features.
        y: iterator that returns for each element, returns 1 or many classes /
           regression values.
        n_classes: indicator of how many classes the target has.
        batch_size: Mini batch size to accumulate.
        random_state: random state for RNG. Note that it will mutate so use a int value
            for this if you want consistent sized batches.

    Attributes:
        X: input features.
        y: input target.
        n_classes: number of classes.
        batch_size: mini batch size to accumulate.
        input_shape: shape of the input.
        output_shape: shape of the output.
        input_dtype: dtype of input.
        output_dtype: dtype of output.
    """
    def __init__(self, X, y, n_classes, batch_size, random_state=None):
        import dask.dataframe as dd
        # TODO: check X and y dtypes in dask_io like pandas
        self.X = X
        self.y = y
        # save column names
        self.X_columns = list(X.columns)
        if isinstance(y.columns[0], str):
            self.y_columns = list(y.columns)
        else:
            # deal with cases where two DFs have overlapped default numeric colnames
            self.y_columns = len(self.X_columns)+1
            self.y = self.y.rename(columns={y.columns[0]: self.y_columns})
        # combine into a data frame
        self.df = dd.multi.concat([self.X, self.y], axis=1)
        self.n_classes = n_classes

        X_count = X.count().compute()[0]
        X_shape = (X_count, len(self.X.columns))
        y_shape = (X_count, len(self.y.columns))
        self.sample_fraction = batch_size/float(X_count)
        self.input_shape, self.output_shape = _get_in_out_shape(
            X_shape, y_shape, n_classes, batch_size)
        # self.X.dtypes[0], self.y.dtypes[self.y_columns]
        self.input_dtype, self.output_dtype = np.float32, np.float32
        if random_state is None:
            self.random_state = 66
        else:
            self.random_state = random_state
        self.batch_size = batch_size

    def get_feed_params(self):
        """Function returns a dict with data feed params while training.
        Returns:
            A dict with data feed params while training.
        """
        return {'batch_size': self.batch_size}

    def get_feed_dict_fn(self, input_placeholder, output_placeholder):
        """Returns a function, that will sample data and provide it to given
        placeholders.

        Args:
            input_placeholder: tf.Placeholder for input features mini batch.
            output_placeholder: tf.Placeholder for output targets.
        Returns:
            A function that when called samples a random subset of batch size
            from X and y.
        """
        def _feed_dict_fn():
            # TODO: option for with/without replacement (dev version of dask)
            sample = self.df.random_split([self.sample_fraction, 1-self.sample_fraction],
                                          random_state=self.random_state)
            inp = extract_pandas_matrix(sample[0][self.X_columns].compute()).tolist()
            out = extract_pandas_matrix(sample[0][self.y_columns].compute())
            # convert to correct dtype
            inp = np.array(inp, dtype=self.input_dtype)
            # one-hot encode out for each class for cross entropy loss
            if HAS_PANDAS:
                import pandas as pd
                if not isinstance(out, pd.Series):
                    out = out.flatten()
            out_max = self.y.max().compute().values[0]
            encoded_out = np.zeros((out.size, out_max+1), dtype=self.output_dtype)
            encoded_out[np.arange(out.size), out] = 1
            return {input_placeholder.name: inp, output_placeholder.name: encoded_out}
        return _feed_dict_fn



#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import random

from sklearn import datasets
from sklearn.grid_search import GridSearchCV
from sklearn.metrics import accuracy_score, mean_squared_error

import tensorflow as tf

import skflow



class GridSearchTest(tf.test.TestCase):

    def testIrisDNN(self):
        random.seed(42)
        iris = datasets.load_iris()
        classifier = skflow.TensorFlowDNNClassifier(
            hidden_units=[10, 20, 10], n_classes=3, steps=50)
        grid_search = GridSearchCV(classifier,
            {'hidden_units': [[5, 5], [10, 10]],
             'learning_rate': [0.1, 0.01]})
        grid_search.fit(iris.data, iris.target)
        score = accuracy_score(iris.target, grid_search.predict(iris.data))
        self.assertGreater(score, 0.5, "Failed with score = {0}".format(score))


if __name__ == "__main__":
    tf.test.main()



#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from struct import Struct
import numpy as np

import tensorflow as tf

from skflow import data_feeder
from skflow.io import *


class MockPlaceholder(object):

    def __init__(self, name):
        self.name = name


class DataFeederTest(tf.test.TestCase):

    def test_data_feeder_regression(self):
        X = np.matrix([[1, 2], [3, 4]])
        y = np.array([1, 2])
        df = data_feeder.DataFeeder(X, y, n_classes=0, batch_size=3)
        feed_dict_fn = df.get_feed_dict_fn(
            MockPlaceholder(name='input'),
            MockPlaceholder(name='output'))
        feed_dict = feed_dict_fn()

        self.assertAllClose(feed_dict['input'], [[3, 4], [1, 2]])
        self.assertAllClose(feed_dict['output'], [2, 1])

    def test_data_feeder_multioutput_regression(self):
        X = np.matrix([[1, 2], [3, 4]])
        y = np.array([[1, 2], [3, 4]])
        df = data_feeder.DataFeeder(X, y, n_classes=0, batch_size=2)
        feed_dict_fn = df.get_feed_dict_fn(
            MockPlaceholder(name='input'),
            MockPlaceholder(name='output'))
        feed_dict = feed_dict_fn()
        self.assertAllClose(feed_dict['input'], [[3, 4], [1, 2]])
        self.assertAllClose(feed_dict['output'], [[3, 4], [1, 2]])

    def test_data_feeder_multioutput_classification(self):
        X = np.matrix([[1, 2], [3, 4]])
        y = np.array([[0, 1, 2], [2, 3, 4]])
        df = data_feeder.DataFeeder(X, y, n_classes=5, batch_size=2)
        feed_dict_fn = df.get_feed_dict_fn(
            MockPlaceholder(name='input'),
            MockPlaceholder(name='output'))
        feed_dict = feed_dict_fn()
        self.assertAllClose(feed_dict['input'], [[3, 4], [1, 2]])
        self.assertAllClose(feed_dict['output'], [[[0, 0, 1, 0, 0],
                                                   [0, 0, 0, 1, 0],
                                                   [0, 0, 0, 0, 1]],
                                                  [[1, 0, 0, 0, 0],
                                                   [0, 1, 0, 0, 0],
                                                   [0, 0, 1, 0, 0]]])

    def test_streaming_data_feeder(self):
        def X_iter():
            yield np.array([1, 2])
            yield np.array([3, 4])

        def y_iter():
            yield np.array([1])
            yield np.array([2])
        df = data_feeder.StreamingDataFeeder(X_iter(), y_iter(), n_classes=0,
                                             batch_size=2)
        feed_dict_fn = df.get_feed_dict_fn(
            MockPlaceholder(name='input'),
            MockPlaceholder(name='output'))
        feed_dict = feed_dict_fn()
        self.assertAllClose(feed_dict['input'], [[1, 2], [3, 4]])
        self.assertAllClose(feed_dict['output'], [1, 2])

    def test_dask_data_feeder(self):
        if HAS_PANDAS and HAS_DASK:
            X = pd.DataFrame(dict(a=np.array([.1, .3, .4, .6, .2, .1, .6]),
                                  b=np.array([.7, .8, .1, .2, .5, .3, .9])))
            X = dd.from_pandas(X, npartitions=2)
            y = pd.DataFrame(dict(labels=np.array([1, 0, 2, 1, 0, 1, 2])))
            y = dd.from_pandas(y, npartitions=2)
            # X = extract_dask_data(X)
            # y = extract_dask_labels(y)
            df = data_feeder.DaskDataFeeder(X, y, n_classes=2, batch_size=2)
            feed_dict_fn = df.get_feed_dict_fn(
                MockPlaceholder(name='input'),
                MockPlaceholder(name='output'))
            feed_dict = feed_dict_fn()
            self.assertAllClose(feed_dict['input'], [[ 0.40000001, 0.1],
                                                     [ 0.60000002, 0.2]])
            self.assertAllClose(feed_dict['output'], [[ 0., 0., 1.],
                                                     [ 0., 1., 0.]])

if __name__ == '__main__':
    tf.test.main()



"""Recurrent Neural Network estimators."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function, absolute_import

from sklearn.base import ClassifierMixin, RegressorMixin

from skflow.estimators.base import TensorFlowEstimator
from skflow import models


def null_input_op_fn(X):
    """This function does no transformation on the inputs, used as default"""
    return X

class TensorFlowRNNClassifier(TensorFlowEstimator, ClassifierMixin):
    """TensorFlow RNN Classifier model.

    Parameters:
        rnn_size: The size for rnn cell, e.g. size of your word embeddings.
        cell_type: The type of rnn cell, including rnn, gru, and lstm.
        num_layers: The number of layers of the rnn model.
        input_op_fn: Function that will transform the input tensor, such as
                     creating word embeddings, byte list, etc. This takes
                     an argument X for input and returns transformed X.
        bidirectional: boolean, Whether this is a bidirectional rnn.
        sequence_length: If sequence_length is provided, dynamic calculation is performed.
                 This saves computational time when unrolling past max sequence length.
        initial_state: An initial state for the RNN. This must be a tensor of appropriate type
                       and shape [batch_size x cell.state_size].
        n_classes: Number of classes in the target.
        tf_master: TensorFlow master. Empty string is default for local.
        batch_size: Mini batch size.
        steps: Number of steps to run over data.
        optimizer: Optimizer name (or class), for example "SGD", "Adam",
                   "Adagrad".
        learning_rate: If this is constant float value, no decay function is used.
            Instead, a customized decay function can be passed that accepts
            global_step as parameter and returns a Tensor.
            e.g. exponential decay function:
            def exp_decay(global_step):
                return tf.train.exponential_decay(
                    learning_rate=0.1, global_step,
                    decay_steps=2, decay_rate=0.001)
        class_weight: None or list of n_classes floats. Weight associated with
                     classes for loss computation. If not given, all classes are suppose to have
                     weight one.
        tf_random_seed: Random seed for TensorFlow initializers.
            Setting this value, allows consistency between reruns.
        continue_training: when continue_training is True, once initialized
            model will be continuely trained on every call of fit.
        num_cores: Number of cores to be used. (default: 4)
        early_stopping_rounds: Activates early stopping if this is not None.
            Loss needs to decrease at least every every <early_stopping_rounds>
            round(s) to continue training. (default: None)
        max_to_keep: The maximum number of recent checkpoint files to keep.
            As new files are created, older files are deleted.
            If None or 0, all checkpoint files are kept.
            Defaults to 5 (that is, the 5 most recent checkpoint files are kept.)
        keep_checkpoint_every_n_hours: Number of hours between each checkpoint
            to be saved. The default value of 10,000 hours effectively disables the feature.
     """

    def __init__(self, rnn_size, n_classes, cell_type='gru', num_layers=1,
                 input_op_fn=null_input_op_fn,
                 initial_state=None, bidirectional=False,
                 sequence_length=None, tf_master="", batch_size=32,
                 steps=50, optimizer="SGD", learning_rate=0.1,
                 class_weight=None,
                 tf_random_seed=42, continue_training=False,
                 num_cores=4, verbose=1, early_stopping_rounds=None,
                 max_to_keep=5, keep_checkpoint_every_n_hours=10000):
        self.rnn_size = rnn_size
        self.cell_type = cell_type
        self.input_op_fn = input_op_fn
        self.bidirectional = bidirectional
        self.num_layers = num_layers
        self.sequence_length = sequence_length
        self.initial_state = initial_state
        super(TensorFlowRNNClassifier, self).__init__(
            model_fn=self._model_fn,
            n_classes=n_classes, tf_master=tf_master,
            batch_size=batch_size, steps=steps, optimizer=optimizer,
            learning_rate=learning_rate, class_weight=class_weight,
            tf_random_seed=tf_random_seed,
            continue_training=continue_training, num_cores=num_cores,
            verbose=verbose,
            early_stopping_rounds=early_stopping_rounds,
            max_to_keep=max_to_keep,
            keep_checkpoint_every_n_hours=keep_checkpoint_every_n_hours)

    def _model_fn(self, X, y):
        return models.get_rnn_model(self.rnn_size, self.cell_type,
                                    self.num_layers,
                                    self.input_op_fn, self.bidirectional,
                                    models.logistic_regression,
                                    self.sequence_length,
                                    self.initial_state)(X, y)

    @property
    def bias_(self):
        """Returns bias of the rnn layer."""
        return self.get_tensor_value('logistic_regression/bias:0')

    @property
    def weights_(self):
        """Returns weights of the rnn layer."""
        return self.get_tensor_value('logistic_regression/weights:0')


class TensorFlowRNNRegressor(TensorFlowEstimator, RegressorMixin):
    """TensorFlow RNN Regressor model.

    Parameters:
        rnn_size: The size for rnn cell, e.g. size of your word embeddings.
        cell_type: The type of rnn cell, including rnn, gru, and lstm.
        num_layers: The number of layers of the rnn model.
        input_op_fn: Function that will transform the input tensor, such as
                     creating word embeddings, byte list, etc. This takes
                     an argument X for input and returns transformed X.
        bidirectional: boolean, Whether this is a bidirectional rnn.
        sequence_length: If sequence_length is provided, dynamic calculation is performed.
                 This saves computational time when unrolling past max sequence length.
        initial_state: An initial state for the RNN. This must be a tensor of appropriate type
                       and shape [batch_size x cell.state_size].
        tf_master: TensorFlow master. Empty string is default for local.
        batch_size: Mini batch size.
        steps: Number of steps to run over data.
        optimizer: Optimizer name (or class), for example "SGD", "Adam",
                   "Adagrad".
        learning_rate: If this is constant float value, no decay function is used.
            Instead, a customized decay function can be passed that accepts
            global_step as parameter and returns a Tensor.
            e.g. exponential decay function:
            def exp_decay(global_step):
                return tf.train.exponential_decay(
                    learning_rate=0.1, global_step,
                    decay_steps=2, decay_rate=0.001)
        tf_random_seed: Random seed for TensorFlow initializers.
            Setting this value, allows consistency between reruns.
        continue_training: when continue_training is True, once initialized
            model will be continuely trained on every call of fit.
        num_cores: Number of cores to be used. (default: 4)
        early_stopping_rounds: Activates early stopping if this is not None.
            Loss needs to decrease at least every every <early_stopping_rounds>
            round(s) to continue training. (default: None)
        verbose: Controls the verbosity, possible values:
                 0: the algorithm and debug information is muted.
                 1: trainer prints the progress.
                 2: log device placement is printed.
        early_stopping_rounds: Activates early stopping if this is not None.
            Loss needs to decrease at least every every <early_stopping_rounds>
            round(s) to continue training. (default: None)
        max_to_keep: The maximum number of recent checkpoint files to keep.
            As new files are created, older files are deleted.
            If None or 0, all checkpoint files are kept.
            Defaults to 5 (that is, the 5 most recent checkpoint files are kept.)
        keep_checkpoint_every_n_hours: Number of hours between each checkpoint
            to be saved. The default value of 10,000 hours effectively disables the feature.
   """

    def __init__(self, rnn_size, cell_type='gru', num_layers=1,
                 input_op_fn=null_input_op_fn, initial_state=None,
                 bidirectional=False, sequence_length=None,
                 n_classes=0, tf_master="", batch_size=32,
                 steps=50, optimizer="SGD", learning_rate=0.1,
                 tf_random_seed=42, continue_training=False,
                 num_cores=4, verbose=1, early_stopping_rounds=None,
                 max_to_keep=5, keep_checkpoint_every_n_hours=10000):
        self.rnn_size = rnn_size
        self.cell_type = cell_type
        self.input_op_fn = input_op_fn
        self.bidirectional = bidirectional
        self.num_layers = num_layers
        self.sequence_length = sequence_length
        self.initial_state = initial_state
        super(TensorFlowRNNRegressor, self).__init__(
            model_fn=self._model_fn,
            n_classes=n_classes, tf_master=tf_master,
            batch_size=batch_size, steps=steps, optimizer=optimizer,
            learning_rate=learning_rate, tf_random_seed=tf_random_seed,
            continue_training=continue_training, num_cores=num_cores, verbose=verbose,
            early_stopping_rounds=early_stopping_rounds,
            max_to_keep=max_to_keep,
            keep_checkpoint_every_n_hours=keep_checkpoint_every_n_hours)

    def _model_fn(self, X, y):
        return models.get_rnn_model(self.rnn_size, self.cell_type,
                                    self.num_layers,
                                    self.input_op_fn, self.bidirectional,
                                    models.linear_regression,
                                    self.sequence_length,
                                    self.initial_state)(X, y)

    @property
    def bias_(self):
        """Returns bias of the rnn layer."""
        return self.get_tensor_value('linear_regression/bias:0')

    @property
    def weights_(self):
        """Returns weights of the rnn layer."""
        return self.get_tensor_value('linear_regression/weights:0')



"""TensorFlow ops for deep neural networks."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function, absolute_import

import tensorflow as tf

import skflow


def dnn(tensor_in, hidden_units, activation=tf.nn.relu, keep_prob=None):
    """Creates fully connected deep neural network subgraph.

    Args:
        tenson_in: tensor or placeholder for input features.
        hidden_units: list of counts of hidden units in each layer.
        activation: activation function between layers. Can be None.
        keep_proba: if not None, will add a dropout layer with given
                    probability.

    Returns:
        A tensor which would be a deep neural network.
    """
    with tf.variable_scope('dnn'):
        for i, n_units in enumerate(hidden_units):
            with tf.variable_scope('layer%d' % i):
                tensor_in = tf.nn.rnn_cell.linear(tensor_in, n_units, True)
                if activation:
                    tensor_in = activation(tensor_in)
                if keep_prob:
                    tensor_in = skflow.ops.dropout(tensor_in, keep_prob)
        return tensor_in




#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import tensorflow as tf

import random

from sklearn import datasets, metrics
from sklearn.cross_validation import train_test_split

import skflow

class CustomDecayTest(tf.test.TestCase):

    def testIrisExponentialDecay(self):
        random.seed(42)

        iris = datasets.load_iris()
        X_train, X_test, y_train, y_test = train_test_split(iris.data,
                                                            iris.target,
                                                            test_size=0.2,
                                                            random_state=42)
        # setup exponential decay function
        def exp_decay(global_step):
            return tf.train.exponential_decay(
                learning_rate=0.1, global_step=global_step,
                decay_steps=100, decay_rate=0.001)
        classifier = skflow.TensorFlowDNNClassifier(hidden_units=[10, 20, 10],
                                                    n_classes=3, steps=800,
                                                    learning_rate=exp_decay)
        classifier.fit(X_train, y_train)
        score = metrics.accuracy_score(y_test, classifier.predict(X_test))

        self.assertGreater(score, 0.7, "Failed with score = {0}".format(score))

if __name__ == "__main__":
    tf.test.main()



"""Generic trainer for TensorFlow models."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function, absolute_import

import sys
import math
from six.moves import xrange   # pylint: disable=redefined-builtin

import numpy as np
import tensorflow as tf


OPTIMIZER_CLS_NAMES = {
    "SGD": tf.train.GradientDescentOptimizer,
    "Adagrad": tf.train.AdagradOptimizer,
    "Adam": tf.train.AdamOptimizer,
}


def _print_report(print_loss_buffer, global_step, epoch):
    """Prints report for given losses and global step."""
    avg_loss = np.mean(print_loss_buffer)
    if epoch:
        print("Step #{step}, epoch #{epoch}, avg. loss: {loss:.5f}"
              .format(step=global_step, loss=avg_loss, epoch=epoch))
    else:
        print("Step #{step}, avg. loss: {loss:.5f}"
              .format(step=global_step, loss=avg_loss))


class TensorFlowTrainer(object):
    """General trainer class.

    Attributes:
      model: Model object.
      gradients: Gradients tensor.
    """

    def __init__(self, loss, global_step, optimizer,
                 learning_rate, clip_gradients=5.0):
        """Build a trainer part of graph.

        Args:
          loss: Tensor that evaluates to model's loss.
          global_step: Tensor with global step of the model.
          optimizer: Name of the optimizer class (SGD, Adam, Adagrad) or class.
          learning_rate: If this is constant float value, no decay function is used.
                         Instead, a customized decay function can be passed that accepts
                         global_step as parameter and returns a Tensor.
                         e.g. exponential decay function:
                         def exp_decay(global_step):
                            return tf.train.exponential_decay(
                                learning_rate=0.1, global_step=global_step,
                                decay_steps=2, decay_rate=0.001)
        Raises:
            ValueError: if learning_rate is not a float or a callable.
        """
        self.loss = loss
        self.global_step = global_step
        # pylint: disable=redefined-variable-type
        if isinstance(learning_rate, float):
            self._learning_rate = tf.get_variable(
                "learning_rate",
                [],
                initializer=tf.constant_initializer(learning_rate))
        elif callable(learning_rate):
            self._learning_rate = learning_rate(self.global_step)
        else:
            raise ValueError("learning_rate should be a float or a callable function.")
        params = tf.trainable_variables()
        self.gradients = tf.gradients(loss, params)
        if clip_gradients > 0.0:
            self.gradients, self.gradients_norm = tf.clip_by_global_norm(
                self.gradients, clip_gradients)
        grads_and_vars = zip(self.gradients, params)
        if isinstance(optimizer, str):
            self._optimizer = OPTIMIZER_CLS_NAMES[
                optimizer](self._learning_rate)
        else:
            self._optimizer = optimizer(self.learning_rate)
        self.trainer = self._optimizer.apply_gradients(grads_and_vars,
                                                       global_step=global_step,
                                                       name="train")
        # Update ops during training, e.g. batch_norm_ops
        self.trainer = tf.group(self.trainer, *tf.get_collection('update_ops'))
        # Get all initializers for all trainable variables.
        self._initializers = tf.initialize_all_variables()

    def initialize(self, sess):
        """Initalizes all variables.

        Args:
            sess: Session object.

        Returns:
            Values of initializers.
        """
        return sess.run(self._initializers)

    def train(self, sess, feed_dict_fn, steps,
              summary_writer=None, summaries=None,
              print_steps=0, verbose=1, early_stopping_rounds=None,
              feed_params_fn=None):
        """Trains a model for given number of steps, given feed_dict function.

        Args:
            sess: Session object.
            feed_dict_fn: Function that will return a feed dictionary.
            steps: Number of steps to run.
            summary_writer: SummaryWriter object to use for writing summaries.
            summaries: Joined object of all summaries that should be ran.
            print_steps: Number of steps in between printing cost.
            verbose: Controls the verbosity. If set to 0, the algorithm is muted.
            early_stopping_rounds: Activates early stopping if this is not None.
                Loss needs to decrease at least every every <early_stopping_rounds>
                round(s) to continue training. (default: None)
            feed_params_fn: params about data feeder state (epoch, offset)

        Returns:
            List of losses for each step.
        """
        losses, print_loss_buffer = [], []
        print_steps = (print_steps if print_steps else
                       math.ceil(float(steps) / 10))

        min_loss = float('inf')
        min_loss_i = 0
        if early_stopping_rounds is not None:
            sys.stderr.write("Performing early stopping. \n")

        for step in xrange(steps):
            feed_dict = feed_dict_fn()
            if summaries:
                global_step, loss, summ, _ = sess.run(
                    [self.global_step, self.loss, summaries, self.trainer],
                    feed_dict=feed_dict)
            else:
                global_step, loss, _ = sess.run(
                    [self.global_step, self.loss, self.trainer],
                    feed_dict=feed_dict)

            if early_stopping_rounds is not None:
                if loss < min_loss:
                    min_loss = loss
                    min_loss_i = step
                elif step - min_loss_i >= early_stopping_rounds:
                    sys.stderr.write("Stopping. Best step:\n \
                                     step {} with loss {}\n".format(min_loss_i,
                                                                    min_loss))
                    break

            losses.append(loss)
            print_loss_buffer.append(loss)
            if summaries and summary_writer and summ is not None:
                summary_writer.add_summary(summ, global_step)
            if verbose > 0:
                if step % print_steps == 0:
                    if feed_params_fn:
                        feed_params = feed_params_fn()
                        epoch = feed_params['epoch'] if 'epoch' in feed_params else None
                    _print_report(print_loss_buffer, global_step, epoch)
                    print_loss_buffer = []

        return losses


class RestoredTrainer(TensorFlowTrainer):
    """Trainer class  that takes already existing graph."""

    # pylint: disable=super-init-not-called
    def __init__(self, loss, global_step, trainer):
        self.global_step = global_step
        self.loss = loss
        self.trainer = trainer




#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import numpy as np
import tensorflow as tf

from skflow import ops


class Seq2SeqOpsTest(tf.test.TestCase):

    def test_sequence_classifier(self):
        with self.test_session() as session:
            decoding = [tf.placeholder(tf.float32, [2, 2]) for _ in range(3)]
            labels = [tf.placeholder(tf.float32, [2, 2]) for _ in range(3)]
            sampling_decoding = [tf.placeholder(tf.float32, [2, 2]) for _ in range(3)]
            predictions, loss = ops.sequence_classifier(decoding, labels, sampling_decoding)
            pred, cost = session.run([predictions, loss], feed_dict={
                decoding[0].name: [[0.1, 0.9], [0.7, 0.3]],
                decoding[1].name: [[0.9, 0.1], [0.8, 0.2]],
                decoding[2].name: [[0.5, 0.5], [0.4, 0.6]],
                labels[0].name: [[1, 0], [0, 1]],
                labels[1].name: [[1, 0], [0, 1]],
                labels[2].name: [[1, 0], [0, 1]],
                sampling_decoding[0].name: [[0.1, 0.9], [0.7, 0.3]],
                sampling_decoding[1].name: [[0.9, 0.1], [0.8, 0.2]],
                sampling_decoding[2].name: [[0.5, 0.5], [0.4, 0.6]],
            })
        self.assertAllEqual(pred.argmax(axis=2), [[1, 0, 0], [0, 0, 1]])
        self.assertAllClose(cost, 4.7839908599)

    def test_seq2seq_inputs(self):
        inp = np.array([[[1, 0], [0, 1], [1, 0]], [[0, 1], [1, 0], [0, 1]]])
        out = np.array([[[0, 1], [1, 0]], [[1, 0], [0, 1]]])
        with self.test_session() as session:
            X = tf.placeholder(tf.float32, [2, 3, 2])
            y = tf.placeholder(tf.float32, [2, 2, 2])
            in_X, in_y, out_y = ops.seq2seq_inputs(X, y, 3, 2)
            enc_inp = session.run(in_X, feed_dict={X.name: inp})
            dec_inp = session.run(in_y, feed_dict={X.name: inp, y.name: out})
            dec_out = session.run(out_y, feed_dict={X.name: inp, y.name: out})
        # Swaps from batch x len x height to list of len of batch x height.
        self.assertAllEqual(enc_inp, np.swapaxes(inp, 0, 1))
        self.assertAllEqual(dec_inp, [[[0, 0], [0, 0]], 
                                      [[0, 1], [1, 0]],
                                      [[1, 0], [0, 1]]]) 
        self.assertAllEqual(dec_out, [[[0, 1], [1, 0]],
                                      [[1, 0], [0, 1]],
                                      [[0, 0], [0, 0]]]) 

    def test_rnn_decoder(self):
        with self.test_session() as session:
            decoder_inputs = [
                tf.placeholder(tf.float32, [2, 2]) for _ in range(3)]
            encoding = tf.placeholder(tf.float32, [2, 2])
            cell = tf.nn.rnn_cell.GRUCell(2)
            outputs, states, sampling_outputs, sampling_states = (
                ops.rnn_decoder(decoder_inputs, encoding, cell))
            self.assertEqual(len(outputs), 3)
            self.assertEqual(outputs[0].get_shape(), [2, 2])
            self.assertEqual(len(states), 4)
            self.assertEqual(states[0].get_shape(), [2, 2])
            self.assertEqual(len(sampling_outputs), 3)
            self.assertEqual(sampling_outputs[0].get_shape(), [2, 2])
            self.assertEqual(len(sampling_states), 4)
            self.assertEqual(sampling_states[0].get_shape(), [2, 2])


if __name__ == '__main__':
    tf.test.main()



"""Implements a number of text preprocessing utilities."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function, absolute_import

import re
import six

import numpy as np

from skflow.preprocessing.categorical_vocabulary import CategoricalVocabulary

TOKENIZER_RE = re.compile(
    r"[A-Z]{2,}(?![a-z])|[A-Z][a-z]+(?=[A-Z])|[\'\w\-]+", re.UNICODE)


def tokenizer(iterator):
    """Tokenizer generator.

    Args:
        iterator: Input iterator with strings.

    Yields:
        array of tokens per each value in the input.
    """
    for value in iterator:
        yield TOKENIZER_RE.findall(value)


class ByteProcessor(object):
    """Maps documents into sequence of ids for bytes."""

    def __init__(self, max_document_length):
        self.max_document_length = max_document_length

    def fit(self, X):
        """Does nothing. No fitting required."""
        pass

    def fit_transform(self, X):
        """Calls transform."""
        return self.transform(X)

    # pylint: disable=no-self-use
    def reverse(self, X):
        """Reverses output of transform back to text.

        Args:
            X: iterator or matrix of integers.
               Document representation in bytes.

        Returns:
            Iterators of utf-8 strings.
        """
        for data in X:
            document = np.trim_zeros(data.astype(np.int8), trim='b').tostring()
            try:
                yield document.decode('utf-8')
            except UnicodeDecodeError:
                yield ''

    def transform(self, X):
        """Transforms input documents into sequence of ids.

        Args:
            X: iterator or list of input documents.
               Documents can be bytes or unicode strings, which will be encoded
               as utf-8 to map to bytes. Note, in Python2 str and bytes is the
               same type.
        Returns:
            iterator of byte ids.
        """
        if six.PY3:
            # For Python3 defined buffer as memoryview.
            buffer_or_memoryview = memoryview
        else:
            buffer_or_memoryview = buffer  # pylint: disable=undefined-variable
        for document in X:
            if isinstance(document, six.text_type):
                document = document.encode('utf-8')
            document_mv = buffer_or_memoryview(document)
            buff = np.frombuffer(document_mv[:self.max_document_length],
                                 dtype=np.uint8)
            yield np.pad(buff, (0, self.max_document_length - len(buff)),
                         'constant')


class VocabularyProcessor(object):
    """Maps documents to sequences of word ids.

    Parameters:
        max_document_length: Maximum length of documents.
            if documents are longer, they will be trimmed, if shorter - padded.
        min_frequency: Minimum frequency of words in the vocabulary.
        vocabulary: CategoricalVocabulary object.

    Attributes:
        vocabulary_: CategoricalVocabulary object.
    """

    def __init__(self, max_document_length,
                 min_frequency=0, vocabulary=None,
                 tokenizer_fn=None):
        self.max_document_length = max_document_length
        self.min_frequency = min_frequency
        if vocabulary:
            self.vocabulary_ = vocabulary
        else:
            self.vocabulary_ = CategoricalVocabulary()
        if tokenizer_fn:
            self._tokenizer = tokenizer_fn
        else:
            self._tokenizer = tokenizer

    def fit(self, raw_documents, unused_y=None):
        """Learn a vocabulary dictionary of all tokens in the raw documents.

        Args:
            raw_documents: iterable
                An iterable which yield either str or unicode.
            unused_y: to match fit format signature of estimators.

        Returns:
            self
        """
        for tokens in self._tokenizer(raw_documents):
            for token in tokens:
                self.vocabulary_.add(token)
        if self.min_frequency > 0:
            self.vocabulary_.trim(self.min_frequency)
        self.vocabulary_.freeze()
        return self

    def fit_transform(self, raw_documents, unused_y=None):
        """Learn the vocabulary dictionary and return indexies of words.

        Args:
            raw_documents: iterable
                An iterable which yield either str or unicode.
            unused_y: to match fit_transform signature of estimators.

        Returns:
            X: iterable, [n_samples, max_document_length]
                Word-id matrix.
        """
        self.fit(raw_documents)
        return self.transform(raw_documents)

    def transform(self, raw_documents):
        """Transform documents to word-id matrix.

        Convert words to ids with vocabulary fitted with fit or the one
        provided in the constructor.

        Args:
            raw_documents: iterable.
                An iterable which yield either str or unicode.

        Returns:
            X: iterable, [n_samples, max_document_length]
                Word-id matrix.
        """
        for tokens in self._tokenizer(raw_documents):
            word_ids = np.zeros(self.max_document_length, np.int64)
            for idx, token in enumerate(tokens):
                if idx >= self.max_document_length:
                    break
                word_ids[idx] = self.vocabulary_.get(token)
            yield word_ids

    def reverse(self, documents):
        """Reverses output of vocabulary mapping to words.

        Args:
            documents: iterable, list of class ids.

        Returns:
            Iterator over mapped in words documents.
        """
        for item in documents:
            output = []
            for class_id in item:
                output.append(self.vocabulary_.reverse(class_id))
            yield ' '.join(output)




"""Deep Neural Network estimators."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function, absolute_import

from sklearn.base import ClassifierMixin, RegressorMixin

from skflow.estimators.base import TensorFlowEstimator
from skflow import models


class TensorFlowDNNClassifier(TensorFlowEstimator, ClassifierMixin):
    """TensorFlow DNN Classifier model.

    Parameters:
        hidden_units: List of hidden units per layer.
        n_classes: Number of classes in the target.
        tf_master: TensorFlow master. Empty string is default for local.
        batch_size: Mini batch size.
        steps: Number of steps to run over data.
        optimizer: Optimizer name (or class), for example "SGD", "Adam",
                   "Adagrad".
        learning_rate: If this is constant float value, no decay function is used.
            Instead, a customized decay function can be passed that accepts
            global_step as parameter and returns a Tensor.
            e.g. exponential decay function:
            def exp_decay(global_step):
                return tf.train.exponential_decay(
                    learning_rate=0.1, global_step,
                    decay_steps=2, decay_rate=0.001)
        class_weight: None or list of n_classes floats. Weight associated with
                     classes for loss computation. If not given, all classes are suppose to have
                     weight one.
        tf_random_seed: Random seed for TensorFlow initializers.
            Setting this value, allows consistency between reruns.
        continue_training: when continue_training is True, once initialized
            model will be continuely trained on every call of fit.
        num_cores: Number of cores to be used. (default: 4)
        early_stopping_rounds: Activates early stopping if this is not None.
            Loss needs to decrease at least every every <early_stopping_rounds>
            round(s) to continue training. (default: None)
        max_to_keep: The maximum number of recent checkpoint files to keep.
            As new files are created, older files are deleted.
            If None or 0, all checkpoint files are kept.
            Defaults to 5 (that is, the 5 most recent checkpoint files are kept.)
        keep_checkpoint_every_n_hours: Number of hours between each checkpoint
            to be saved. The default value of 10,000 hours effectively disables the feature.
     """

    def __init__(self, hidden_units, n_classes, tf_master="", batch_size=32,
                 steps=200, optimizer="SGD", learning_rate=0.1,
                 class_weight=None,
                 tf_random_seed=42, continue_training=False, num_cores=4,
                 verbose=1, early_stopping_rounds=None,
                 max_to_keep=5, keep_checkpoint_every_n_hours=10000):
        self.hidden_units = hidden_units
        super(TensorFlowDNNClassifier, self).__init__(
            model_fn=self._model_fn,
            n_classes=n_classes, tf_master=tf_master,
            batch_size=batch_size, steps=steps, optimizer=optimizer,
            learning_rate=learning_rate, class_weight=class_weight,
            tf_random_seed=tf_random_seed,
            continue_training=continue_training, num_cores=4, verbose=verbose,
            early_stopping_rounds=early_stopping_rounds,
            max_to_keep=max_to_keep,
            keep_checkpoint_every_n_hours=keep_checkpoint_every_n_hours)

    def _model_fn(self, X, y):
        return models.get_dnn_model(self.hidden_units,
                                    models.logistic_regression)(X, y)

    @property
    def weights_(self):
        """Returns weights of the DNN weight layers."""
        weights = []
        for layer in range(len(self.hidden_units)):
            weights.append(self.get_tensor_value('dnn/layer%d/Linear/Matrix:0' % layer))
        weights.append(self.get_tensor_value('logistic_regression/weights:0'))
        return weights

    @property
    def bias_(self):
        """Returns bias of the DNN's bias layers."""
        biases = []
        for layer in range(len(self.hidden_units)):
            biases.append(self.get_tensor_value('dnn/layer%d/Linear/Bias:0' % layer))
        biases.append(self.get_tensor_value('logistic_regression/bias:0'))
        return biases


class TensorFlowDNNRegressor(TensorFlowEstimator, RegressorMixin):
    """TensorFlow DNN Regressor model.

    Parameters:
        hidden_units: List of hidden units per layer.
        tf_master: TensorFlow master. Empty string is default for local.
        batch_size: Mini batch size.
        steps: Number of steps to run over data.
        optimizer: Optimizer name (or class), for example "SGD", "Adam",
                   "Adagrad".
        learning_rate: If this is constant float value, no decay function is used.
            Instead, a customized decay function can be passed that accepts
            global_step as parameter and returns a Tensor.
            e.g. exponential decay function:
            def exp_decay(global_step):
                return tf.train.exponential_decay(
                    learning_rate=0.1, global_step,
                    decay_steps=2, decay_rate=0.001)
        tf_random_seed: Random seed for TensorFlow initializers.
            Setting this value, allows consistency between reruns.
        continue_training: when continue_training is True, once initialized
            model will be continuely trained on every call of fit.
        num_cores: Number of cores to be used. (default: 4)
        early_stopping_rounds: Activates early stopping if this is not None.
            Loss needs to decrease at least every every <early_stopping_rounds>
            round(s) to continue training. (default: None)
        verbose: Controls the verbosity, possible values:
                 0: the algorithm and debug information is muted.
                 1: trainer prints the progress.
                 2: log device placement is printed.
        early_stopping_rounds: Activates early stopping if this is not None.
            Loss needs to decrease at least every every <early_stopping_rounds>
            round(s) to continue training. (default: None)
        max_to_keep: The maximum number of recent checkpoint files to keep.
            As new files are created, older files are deleted.
            If None or 0, all checkpoint files are kept.
            Defaults to 5 (that is, the 5 most recent checkpoint files are kept.)
        keep_checkpoint_every_n_hours: Number of hours between each checkpoint
            to be saved. The default value of 10,000 hours effectively disables the feature.
   """

    def __init__(self, hidden_units, n_classes=0, tf_master="", batch_size=32,
                 steps=200, optimizer="SGD", learning_rate=0.1,
                 tf_random_seed=42, continue_training=False, num_cores=4,
                 verbose=1, early_stopping_rounds=None,
                 max_to_keep=5, keep_checkpoint_every_n_hours=10000):
        self.hidden_units = hidden_units
        super(TensorFlowDNNRegressor, self).__init__(
            model_fn=self._model_fn,
            n_classes=n_classes, tf_master=tf_master,
            batch_size=batch_size, steps=steps, optimizer=optimizer,
            learning_rate=learning_rate, tf_random_seed=tf_random_seed,
            continue_training=continue_training, num_cores=num_cores, verbose=verbose,
            early_stopping_rounds=early_stopping_rounds,
            max_to_keep=max_to_keep,
            keep_checkpoint_every_n_hours=keep_checkpoint_every_n_hours)

    def _model_fn(self, X, y):
        return models.get_dnn_model(self.hidden_units,
                                    models.linear_regression)(X, y)

    @property
    def weights_(self):
        """Returns weights of the DNN weight layers."""
        weights = []
        for layer in range(len(self.hidden_units)):
            weights.append(self.get_tensor_value('dnn/layer%d/Linear/Matrix:0' % layer))
        weights.append(self.get_tensor_value('linear_regression/weights:0'))
        return weights

    @property
    def bias_(self):
        """Returns bias of the DNN's bias layers."""
        biases = []
        for layer in range(len(self.hidden_units)):
            biases.append(self.get_tensor_value('dnn/layer%d/Linear/Bias:0' % layer))
        biases.append(self.get_tensor_value('linear_regression/bias:0'))
        return biases




#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import random

from sklearn import datasets
from sklearn.metrics import accuracy_score, mean_squared_error, log_loss

import tensorflow as tf

import skflow


class SaverTest(tf.test.TestCase):

    def testIris(self):
        path = tf.test.get_temp_dir() + '/tmp.saver'
        random.seed(42)
        iris = datasets.load_iris()
        classifier = skflow.TensorFlowLinearClassifier(n_classes=3)
        classifier.fit(iris.data, iris.target)
        classifier.save(path)
        new_classifier = skflow.TensorFlowEstimator.restore(path)
        self.assertEqual(type(new_classifier), type(classifier))
        score = accuracy_score(iris.target, new_classifier.predict(iris.data))
        self.assertGreater(score, 0.5, "Failed with score = {0}".format(score))

    def testCustomModel(self):
        path = tf.test.get_temp_dir() + '/tmp.saver2'
        random.seed(42)
        iris = datasets.load_iris()
        def custom_model(X, y):
            return skflow.models.logistic_regression(X, y)
        classifier = skflow.TensorFlowEstimator(model_fn=custom_model,
            n_classes=3)
        classifier.fit(iris.data, iris.target)
        classifier.save(path)
        new_classifier = skflow.TensorFlowEstimator.restore(path)
        self.assertEqual(type(new_classifier), type(classifier))
        score = accuracy_score(iris.target, new_classifier.predict(iris.data))
        self.assertGreater(score, 0.5, "Failed with score = {0}".format(score))
    
    def testDNN(self):
        path = tf.test.get_temp_dir() + '/tmp_saver3'
        random.seed(42)
        iris = datasets.load_iris()
        classifier = skflow.TensorFlowDNNClassifier(hidden_units=[10, 20, 10], n_classes=3)
        classifier.fit(iris.data, iris.target)
        classifier.save(path)
        new_classifier = skflow.TensorFlowEstimator.restore(path)
        self.assertEqual(type(new_classifier), type(classifier))
        score = accuracy_score(iris.target, new_classifier.predict(iris.data))
        self.assertGreater(score, 0.5, "Failed with score = {0}".format(score))

    def testNoFolder(self):
        with self.assertRaises(ValueError):
            skflow.TensorFlowEstimator.restore('no_model_path')

    def testNoCheckpoints(self):
        path = tf.test.get_temp_dir() + '/tmp/tmp.saver4'
        random.seed(42)
        iris = datasets.load_iris()
        classifier = skflow.TensorFlowDNNClassifier(hidden_units=[10, 20, 10], n_classes=3)
        classifier.fit(iris.data, iris.target)
        classifier.save(path)
        os.remove(os.path.join(path, 'checkpoint'))
        with self.assertRaises(ValueError):
            skflow.TensorFlowEstimator.restore(path)
        

if __name__ == "__main__":
    tf.test.main()



"""Various TensorFlow Ops."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from skflow.ops.array_ops import *
from skflow.ops.conv_ops import *
from skflow.ops.dnn_ops import *
from skflow.ops.dropout_ops import *
from skflow.ops.embeddings_ops import *
from skflow.ops.losses_ops import *
from skflow.ops.seq2seq_ops import *
from skflow.ops.batch_norm_ops import *



"""TensorFlow Ops for loss computation."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function, absolute_import

import tensorflow as tf


def mean_squared_error_regressor(tensor_in, labels, weights, biases, name=None):
    """Returns prediction and loss for mean squared error regression."""
    with tf.op_scope([tensor_in, labels], name, "mean_squared_error_regressor"):
        predictions = tf.nn.xw_plus_b(tensor_in, weights, biases)
        if len(labels.get_shape()) == 1:
            labels = tf.reshape(labels, [-1, 1])
        diff = labels - predictions
        loss = tf.reduce_mean(tf.mul(diff, diff))
        return predictions, loss


def softmax_classifier(tensor_in, labels, weights, biases, class_weight=None, name=None):
    """Returns prediction and loss for softmax classifier.

    Args:
        tensor_in: Input tensor, [batch_size, feature_size], features.
        labels: Tensor, [batch_size, n_classes], labels of the output classes.
        weights: Tensor, [batch_size, feature_size], linear transformation matrix.
        biases: Tensor, [batch_size], biases.
        class_weight: Tensor, optional, [n_classes], weight for each class.
                      If not given, all classes are supposed to have weight
                      one.

    Returns:
        Prediction and loss tensors.
    """
    with tf.op_scope([tensor_in, labels], name, "softmax_classifier"):
        logits = tf.nn.xw_plus_b(tensor_in, weights, biases)
        if class_weight:
            logits = tf.mul(logits, class_weight)
        xent = tf.nn.softmax_cross_entropy_with_logits(logits,
                                                       labels,
                                                       name="xent_raw")
        loss = tf.reduce_mean(xent, name="xent")
        predictions = tf.nn.softmax(logits, name=name)
        return predictions, loss




"""TensorFlow Ops for Sequence to Sequence models."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function, absolute_import

import tensorflow as tf

from skflow.ops import array_ops


def sequence_classifier(decoding, labels, sampling_decoding=None, name=None):
    """Returns predictions and loss for sequence of predictions.

    Args:
        decoding: List of Tensors with predictions.
        labels: List of Tensors with labels.
        sampling_decoding: Optional, List of Tensor with predictions to be used
                           in sampling. E.g. they shouldn't have dependncy on ouptuts.
                           If not provided, decoding is used.

    Returns:
        Predictions and losses tensors.
    """
    with tf.op_scope([decoding, labels], name, "sequence_classifier"):
        predictions, xent_list = [], []
        for i, pred in enumerate(decoding):
            xent_list.append(
                tf.nn.softmax_cross_entropy_with_logits(
                    pred, labels[i], name="sequence_loss/xent_raw{0}".format(i)))
            if sampling_decoding:
                predictions.append(tf.nn.softmax(sampling_decoding[i]))
            else:
                predictions.append(tf.nn.softmax(pred))
        xent = tf.add_n(xent_list, name="sequence_loss/xent")
        loss = tf.reduce_sum(xent, name="sequence_loss")
        return array_ops.expand_concat(1, predictions), loss


def seq2seq_inputs(X, y, input_length, output_length, sentinel=None, name=None):
    """Processes inputs for Sequence to Sequence models.

    Args:
        X: Input Tensor [batch_size, input_length, embed_dim].
        y: Output Tensor [batch_size, output_length, embed_dim].
        input_length: length of input X.
        output_length: length of output y.
        sentinel: optional first input to decoder and final output expected.
                  if sentinel is not provided, zeros are used.
                  Due to fact that y is not available in sampling time, shape
                  of sentinel will be inferred from X.

    Returns:
        Encoder input from X, and decoder inputs and outputs from y.
    """
    with tf.op_scope([X, y], name, "seq2seq_inputs"):
        in_X = array_ops.split_squeeze(1, input_length, X)
        y = array_ops.split_squeeze(1, output_length, y)
        if not sentinel:
            # Set to zeros of shape of X[0]
            sentinel = tf.zeros(tf.shape(in_X[0]))
            sentinel.set_shape(in_X[0].get_shape())
        in_y = [sentinel] + y
        out_y = y + [sentinel]
        return in_X, in_y, out_y


def rnn_decoder(decoder_inputs, initial_state, cell, scope=None):
    """RNN Decoder that creates training and sampling sub-graphs.

    Args:
        decoder_inputs: Inputs for decoder, list of tensors.
                        This is used only in trianing sub-graph.
        initial_state: Initial state for the decoder.
        cell: RNN cell to use for decoder.
        scope: Scope to use, if None new will be produced.

    Returns:
        List of tensors for outputs and states for training and sampling sub-graphs.
    """
    with tf.variable_scope(scope or "dnn_decoder"):
        states, sampling_states = [initial_state], [initial_state]
        outputs, sampling_outputs = [], []
        with tf.op_scope([decoder_inputs, initial_state], "training"):
            for i, inp in enumerate(decoder_inputs):
                if i > 0:
                    tf.get_variable_scope().reuse_variables()
                output, new_state = cell(inp, states[-1])
                outputs.append(output)
                states.append(new_state)
        with tf.op_scope([initial_state], "sampling"):
            for i, _ in enumerate(decoder_inputs):
                if i == 0:
                    sampling_outputs.append(outputs[i])
                    sampling_states.append(states[i])
                else:
                    sampling_output, sampling_state = cell(
                        sampling_outputs[-1], sampling_states[-1])
                    sampling_outputs.append(sampling_output)
                    sampling_states.append(sampling_state)
    return outputs, states, sampling_outputs, sampling_states


def rnn_seq2seq(encoder_inputs, decoder_inputs, cell, dtype=tf.float32, scope=None):
    """RNN Sequence to Sequence model.

    Args:
        encoder_inputs: List of tensors, inputs for encoder.
        decoder_inputs: List of tensors, inputs for decoder.
        cell: RNN cell to use for encoder and decoder.
        dtype: Type to initialize encoder state with.
        scope: Scope to use, if None new will be produced.

    Returns:
        List of tensors for outputs and states for trianing and sampling sub-graphs.
    """
    with tf.variable_scope(scope or "rnn_seq2seq"):
        _, enc_states = tf.nn.rnn(cell, encoder_inputs, dtype=dtype)
        return rnn_decoder(decoder_inputs, enc_states[-1], cell)




#!/usr/bin/env python
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from setuptools import setup, find_packages

META_DATA = dict(
    name='skflow',
    version='0.1.0',
    url='https://github.com/tensorflow/skflow',
    license='Apache-2',
    description='Simplified Interface for TensorFlow for Deep Learning',
    author=['Scikit Flow Authors'],
    author_email='terrytangyuan@Gmail.com',
    packages=find_packages(),
    install_requires=[
        'sklearn',
        'scipy',
        'numpy',
    ],
    classifiers=[
        'Programming Language :: Python',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering'
    ],
    keywords=[
        'Deep Learning',
        'Neural Networks',
        'Google',
        'TensorFlow',
        'Machine Learning'
    ]
)


if __name__ == '__main__':
    setup(**META_DATA)




"""Various high level TF models."""
#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function, absolute_import

import tensorflow as tf

from skflow.ops import mean_squared_error_regressor, softmax_classifier, dnn


def linear_regression(X, y):
    """Creates linear regression TensorFlow subgraph.

    Args:
        X: tensor or placeholder for input features.
        y: tensor or placeholder for target.

    Returns:
        Predictions and loss tensors.
    """
    with tf.variable_scope('linear_regression'):
        tf.histogram_summary('linear_regression.X', X)
        tf.histogram_summary('linear_regression.y', y)
        y_shape = y.get_shape()
        if len(y_shape) == 1:
            output_shape = 1
        else:
            output_shape = y_shape[1]
        weights = tf.get_variable('weights', [X.get_shape()[1], output_shape])
        bias = tf.get_variable('bias', [output_shape])
        tf.histogram_summary('linear_regression.weights', weights)
        tf.histogram_summary('linear_regression.bias', bias)
        return mean_squared_error_regressor(X, y, weights, bias)


def logistic_regression(X, y, class_weight=None):
    """Creates logistic regression TensorFlow subgraph.

    Args:
        X: tensor or placeholder for input features,
           shape should be [batch_size, n_features].
        y: tensor or placeholder for target,
           shape should be [batch_size, n_classes].
        class_weight: tensor, [n_classes], where for each class
                      it has weight of the class. If not provided
                      will check if graph contains tensor `class_weight:0`.
                      If that is not provided either all ones are used.

    Returns:
        Predictions and loss tensors.
    """
    with tf.variable_scope('logistic_regression'):
        tf.histogram_summary('logistic_regression.X', X)
        tf.histogram_summary('logistic_regression.y', y)
        weights = tf.get_variable('weights', [X.get_shape()[1],
                                              y.get_shape()[-1]])
        bias = tf.get_variable('bias', [y.get_shape()[-1]])
        tf.histogram_summary('logistic_regression.weights', weights)
        tf.histogram_summary('logistic_regression.bias', bias)
        # If no class weight provided, try to retrieve one from pre-defined
        # tensor name in the graph.
        if not class_weight:
            try:
                class_weight = tf.get_default_graph().get_tensor_by_name('class_weight:0')
            except KeyError:
                pass
        return softmax_classifier(X, y, weights, bias,
                                  class_weight=class_weight)


def get_dnn_model(hidden_units, target_predictor_fn):
    """Returns a function that creates a DNN TensorFlow subgraph with given
    params.

    Args:
        hidden_units: List of values of hidden units for layers.
        target_predictor_fn: Function that will predict target from input
                             features. This can be logistic regression,
                             linear regression or any other model,
                             that takes X, y and returns predictions and loss tensors.

    Returns:
        A function that creates the subgraph.
    """
    def dnn_estimator(X, y):
        """DNN estimator with target predictor function on top."""
        layers = dnn(X, hidden_units)
        return target_predictor_fn(layers, y)
    return dnn_estimator

## This will be in Tensorflow 0.7.
## TODO(ilblackdragon): Clean this up when it's released


def _reverse_seq(input_seq, lengths):
    """Reverse a list of Tensors up to specified lengths.
    Args:
        input_seq: Sequence of seq_len tensors of dimension (batch_size, depth)
        lengths:   A tensor of dimension batch_size, containing lengths for each
                   sequence in the batch. If "None" is specified, simply reverses
                   the list.
    Returns:
        time-reversed sequence
    """
    if lengths is None:
        return list(reversed(input_seq))

    for input_ in input_seq:
        input_.set_shape(input_.get_shape().with_rank(2))

    # Join into (time, batch_size, depth)
    s_joined = tf.pack(input_seq)

    # Reverse along dimension 0
    s_reversed = tf.reverse_sequence(s_joined, lengths, 0, 1)
    # Split again into list
    result = tf.unpack(s_reversed)
    return result


def bidirectional_rnn(cell_fw, cell_bw, inputs,
                      initial_state_fw=None, initial_state_bw=None,
                      dtype=None, sequence_length=None, scope=None):
    """Creates a bidirectional recurrent neural network.
    Similar to the unidirectional case above (rnn) but takes input and builds
    independent forward and backward RNNs with the final forward and backward
    outputs depth-concatenated, such that the output will have the format
    [time][batch][cell_fw.output_size + cell_bw.output_size]. The input_size of
    forward and backward cell must match. The initial state for both directions
    is zero by default (but can be set optionally) and no intermediate states are
    ever returned -- the network is fully unrolled for the given (passed in)
    length(s) of the sequence(s) or completely unrolled if length(s) is not given.
    Args:
        cell_fw: An instance of RNNCell, to be used for forward direction.
        cell_bw: An instance of RNNCell, to be used for backward direction.
        inputs: A length T list of inputs, each a tensor of shape
          [batch_size, cell.input_size].
        initial_state_fw: (optional) An initial state for the forward RNN.
          This must be a tensor of appropriate type and shape
          [batch_size x cell.state_size].
        initial_state_bw: (optional) Same as for initial_state_fw.
        dtype: (optional) The data type for the initial state.  Required if either
          of the initial states are not provided.
        sequence_length: (optional) An int64 vector (tensor) of size [batch_size],
          containing the actual lengths for each of the sequences.
        scope: VariableScope for the created subgraph; defaults to "BiRNN"
    Returns:
        A set of output `Tensors` where:
          outputs is a length T list of outputs (one for each input), which
          are depth-concatenated forward and backward outputs
    Raises:
        TypeError: If "cell_fw" or "cell_bw" is not an instance of RNNCell.
        ValueError: If inputs is None or an empty list.
    """

    if not isinstance(cell_fw, tf.nn.rnn_cell.RNNCell):
        raise TypeError("cell_fw must be an instance of RNNCell")
    if not isinstance(cell_bw, tf.nn.rnn_cell.RNNCell):
        raise TypeError("cell_bw must be an instance of RNNCell")
    if not isinstance(inputs, list):
        raise TypeError("inputs must be a list")
    if not inputs:
        raise ValueError("inputs must not be empty")

    name = scope or "BiRNN"
    # Forward direction
    with tf.variable_scope(name + "_FW"):
        output_fw, _ = tf.nn.rnn(cell_fw, inputs, initial_state_fw, dtype,
                                 sequence_length)

    # Backward direction
    with tf.variable_scope(name + "_BW"):
        tmp, _ = tf.nn.rnn(cell_bw, _reverse_seq(inputs, sequence_length),
                           initial_state_bw, dtype, sequence_length)
    output_bw = _reverse_seq(tmp, sequence_length)
    # Concat each of the forward/backward outputs
    outputs = [tf.concat(1, [fw, bw])
               for fw, bw in zip(output_fw, output_bw)]

    return outputs

# End of Tensorflow 0.7


def get_rnn_model(rnn_size, cell_type, num_layers, input_op_fn,
                  bidirectional, target_predictor_fn,
                  sequence_length, initial_state):
    """Returns a function that creates a RNN TensorFlow subgraph with given
    params.

    Args:
        rnn_size: The size for rnn cell, e.g. size of your word embeddings.
        cell_type: The type of rnn cell, including rnn, gru, and lstm.
        num_layers: The number of layers of the rnn model.
        input_op_fn: Function that will transform the input tensor, such as
                     creating word embeddings, byte list, etc. This takes
                     an argument X for input and returns transformed X.
        bidirectional: boolean, Whether this is a bidirectional rnn.
        target_predictor_fn: Function that will predict target from input
                             features. This can be logistic regression,
                             linear regression or any other model,
                             that takes X, y and returns predictions and loss tensors.
        sequence_length: If sequence_length is provided, dynamic calculation is performed.
                         This saves computational time when unrolling past max sequence length.
                         Required for bidirectional RNNs.
        initial_state: An initial state for the RNN. This must be a tensor of appropriate type
                       and shape [batch_size x cell.state_size].

    Returns:
        A function that creates the subgraph.
    """
    def rnn_estimator(X, y):
        """RNN estimator with target predictor function on top."""
        X = input_op_fn(X)
        if cell_type == 'rnn':
            cell_fn = tf.nn.rnn_cell.BasicRNNCell
        elif cell_type == 'gru':
            cell_fn = tf.nn.rnn_cell.GRUCell
        elif cell_type == 'lstm':
            cell_fn = tf.nn.rnn_cell.BasicLSTMCell
        else:
            raise ValueError("cell_type {} is not supported. ".format(cell_type))
        if bidirectional:
            # forward direction cell
            rnn_fw_cell = tf.nn.rnn_cell.MultiRNNCell([cell_fn(rnn_size)] * num_layers)
            # backward direction cell
            rnn_bw_cell = tf.nn.rnn_cell.MultiRNNCell([cell_fn(rnn_size)] * num_layers)
            # pylint: disable=unexpected-keyword-arg, no-value-for-parameter
            encoding = bidirectional_rnn(rnn_fw_cell, rnn_bw_cell, X,
                                         dtype=tf.float32,
                                         sequence_length=sequence_length,
                                         initial_state_fw=initial_state,
                                         initial_state_bw=initial_state)
        else:
            cell = tf.nn.rnn_cell.MultiRNNCell([cell_fn(rnn_size)] * num_layers)
            _, encoding = tf.nn.rnn(cell, X, dtype=tf.float32,
                                    sequence_length=sequence_length,
                                    initial_state=initial_state)
        return target_predictor_fn(encoding[-1], y)
    return rnn_estimator



#  Copyright 2015-present Scikit Flow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import random

from sklearn import datasets
from sklearn.metrics import accuracy_score, mean_squared_error, log_loss

import numpy as np
import tensorflow as tf

import skflow


class RegressionTest(tf.test.TestCase):

    def testLinearRegression(self):
        rng = np.random.RandomState(67)
        N = 1000
        n_weights = 10
        self.bias = 2
        self.X = rng.uniform(-1, 1, (N, n_weights))
        self.weights = 10 * rng.randn(n_weights)
        self.y = np.dot(self.X, self.weights)
        self.y += rng.randn(len(self.X)) * 0.05 + rng.normal(self.bias, 0.01)
        regressor = skflow.TensorFlowLinearRegressor()
        regressor.fit(self.X, self.y)
        # Have to flatten weights since they come in (X, 1) shape
        self.assertAllClose(self.weights, regressor.weights_.flatten(), rtol=0.01)
        assert abs(self.bias - regressor.bias_) < 0.1


if __name__ == "__main__":
    tf.test.main()
